import calendar
from datetime import datetime, timedelta

from django.test import TestCase
from django.urls import reverse

from NEMO.models import User, Tool, ScheduledOutage, Reservation, Account, Project
from NEMO.tests.test_utilities import login_as_user, login_as


class ReservationTestCase(TestCase):
	tool: Tool = None
	owner: User = None
	consumer: User = None
	staff: User = None

	def setUp(self):
		global tool, consumer, staff
		owner = User.objects.create(username='mctest', first_name='Testy', last_name='McTester')
		tool = Tool.objects.create(name='test_tool', primary_owner=owner, _category='Imaging')
		account = Account.objects.create(name="account1")
		project = Project.objects.create(name="project1", account=account)
		staff = User.objects.create(username='staff', first_name='Staff', last_name='Member', is_staff=True)
		consumer = User.objects.create(username='jsmith', first_name='John', last_name='Smith', training_required=False)
		consumer.qualifications.add(tool)
		consumer.projects.add(project)
		consumer.save()

	@staticmethod
	def get_reservation_data(start: datetime, end:datetime, tool_param: Tool):
		return {
			'start':calendar.timegm(start.utctimetuple()),
			'end':calendar.timegm(end.utctimetuple()),
			'tool_name':tool_param.name,
		}

	def test_user_does_not_meet_conditions(self):
		user = User.objects.create(username='noproj', first_name='scott', last_name='NoProj')

		start = datetime.now() + timedelta(hours=1)
		end = start + timedelta(hours=1)
		data = self.get_reservation_data(start, end, tool)

		login_as(self.client, user)

		response = self.client.post(reverse('create_reservation'), data, follow=True)
		self.assertEquals(response.status_code, 200)
		self.assertTrue("You do not belong to any active projects. Thus, you may not create any reservations." in response.content.decode())
		self.assertTrue("You are blocked from making reservations for all tools in the NanoFab. Please complete the NanoFab rules tutorial in order to create new reservations." in response.content.decode())
		self.assertTrue("You are not qualified to use this tool. Creating, moving, and resizing reservations is forbidden." in response.content.decode())

		user.training_required = False
		user.save()
		login_as(self.client, user)
		response = self.client.post(reverse('create_reservation'), data, follow=True)
		self.assertEquals(response.status_code, 200)
		self.assertTrue("You are blocked from making reservations for all tools in the NanoFab. Please complete the NanoFab rules tutorial in order to create new reservations." not in response.content.decode())
		self.assertTrue("You do not belong to any active projects. Thus, you may not create any reservations." in response.content.decode())
		self.assertTrue("You are not qualified to use this tool. Creating, moving, and resizing reservations is forbidden." in response.content.decode())

		user.qualifications.add(tool)
		login_as(self.client, user)
		response = self.client.post(reverse('create_reservation'), data, follow=True)
		self.assertEquals(response.status_code, 200)
		self.assertTrue("You are not qualified to use this tool. Creating, moving, and resizing reservations is forbidden." not in response.content.decode())
		self.assertTrue("You do not belong to any active projects. Thus, you may not create any reservations." in response.content.decode())

	def test_reservation_policy_problems(self):
		start = datetime.now() + timedelta(hours=1)
		end = start - timedelta(hours=1)
		data = self.get_reservation_data(start, end, tool)

		login_as(self.client, consumer)

		response = self.client.post(reverse('create_reservation'), data, follow=True)
		self.assertEquals(response.status_code, 400)
		self.assertEquals(response.content.decode(), "The request parameters have an end time that precedes the start time.")

		# fix time
		end = (start + timedelta(hours=1))
		# Create a outage and try to schedule a reservation at the same time
		outage = ScheduledOutage.objects.create(title="Outage", tool=tool, start=start, end=end, creator=staff)
		data = self.get_reservation_data(start, end, tool)
		response = self.client.post(reverse('create_reservation'), data, follow=True)
		self.assertEquals(response.status_code, 200)
		self.assertTrue("Your reservation coincides with a scheduled outage. Please choose a different time." in response.content.decode())

		# try to schedule a reservation that starts before but ends slightly after the outage starts
		data = self.get_reservation_data(start - timedelta(hours=1), end - timedelta(minutes=59), tool)
		response = self.client.post(reverse('create_reservation'), data, follow=True)
		self.assertEquals(response.status_code, 200)
		self.assertTrue("Your reservation coincides with a scheduled outage. Please choose a different time." in response.content.decode())

		# try to schedule a reservation that starts slightly before the outage ends
		data = self.get_reservation_data(start + timedelta(minutes=59), end + timedelta(hours=1), tool)
		response = self.client.post(reverse('create_reservation'), data, follow=True)
		self.assertEquals(response.status_code, 200)
		self.assertTrue("Your reservation coincides with a scheduled outage. Please choose a different time." in response.content.decode())

		outage.delete()

		# try to schedule a reservation in the past
		data = self.get_reservation_data(start - timedelta(days=1), end - timedelta(days=1), tool)
		response = self.client.post(reverse('create_reservation'), data, follow=True)
		self.assertEquals(response.status_code, 200)
		self.assertTrue("start time" in response.content.decode() and  "is earlier than the current time" in response.content.decode())

		# check tool horizon (days in advance to reserve tool)
		tool.reservation_horizon = 2
		tool.save()
		start = datetime.now() + timedelta(days=3)
		end = start + timedelta(hours=1)
		data = self.get_reservation_data(start, end, tool)

		login_as(self.client, consumer)
		response = self.client.post(reverse('create_reservation'), data, follow=True)
		self.assertEquals(response.status_code, 200)
		self.assertTrue("You may not create reservations further than 2 days from now for this tool." in response.content.decode())
		self.assertEquals(Reservation.objects.filter(tool=tool).count(), 0)

		# minimum & maximum duration
		tool.minimum_usage_block_time = 90
		tool.maximum_usage_block_time = 30
		tool.save()
		start = datetime.now() + timedelta(hours=1)
		end = start + timedelta(hours=1)
		data = self.get_reservation_data(start, end, tool)

		login_as(self.client, consumer)
		response = self.client.post(reverse('create_reservation'), data, follow=True)
		self.assertEquals(response.status_code, 200)
		self.assertTrue("Your reservation has a duration of 60 minutes. This tool requires a minimum reservation duration of 90 minutes." in response.content.decode())
		self.assertTrue("Your reservation has a duration of 60 minutes. Reservations for this tool may not exceed 30 minutes." in response.content.decode())
		self.assertEquals(Reservation.objects.filter(tool=tool).count(), 0)

		# max reservations per day
		first_of_the_day = Reservation.objects.create(tool=tool, start=start, end=end, creator=consumer, user=consumer, short_notice=False)
		tool.maximum_reservations_per_day = 1
		tool.minimum_usage_block_time = None
		tool._maximum_usage_block_time = None
		tool.save()
		data = self.get_reservation_data(start + timedelta(hours=2), end + timedelta(hours=2), tool)
		response = self.client.post(reverse('create_reservation'), data, follow=True)
		self.assertEquals(response.status_code, 200)
		self.assertTrue("You may only have 1 reservations for this tool per day. Missed reservations are included when counting the number of reservations per day" in response.content.decode())

		# even if the first one was missed, still counts towards the limit
		first_of_the_day.missed = True
		first_of_the_day.save()
		data = self.get_reservation_data(start + timedelta(hours=2), end + timedelta(hours=2), tool)
		response = self.client.post(reverse('create_reservation'), data, follow=True)
		self.assertEquals(response.status_code, 200)
		self.assertTrue("You may only have 1 reservations for this tool per day. Missed reservations are included when counting the number of reservations per day" in response.content.decode())

		# test minimum time between reservations
		first_of_the_day.missed = False
		first_of_the_day.save()

		tool.maximum_reservations_per_day = None
		tool.minimum_time_between_reservations = 120
		tool.save()
		data = self.get_reservation_data(start + timedelta(minutes=90), end + timedelta(minutes=90), tool)
		response = self.client.post(reverse('create_reservation'), data, follow=True)
		self.assertEquals(response.status_code, 200)
		self.assertTrue("Separate reservations for this tool that belong to you must be at least 120 minutes apart from each other. The proposed reservation ends too close to another reservation." in response.content.decode())

		data = self.get_reservation_data(start + timedelta(minutes=30), start + timedelta(minutes=90), tool)
		response = self.client.post(reverse('create_reservation'), data, follow=True)
		self.assertEquals(response.status_code, 200)
		self.assertTrue("Separate reservations for this tool that belong to you must be at least 120 minutes apart from each other. The proposed reservation begins too close to another reservation." in response.content.decode())
		self.assertTrue("Separate reservations for this tool that belong to you must be at least 120 minutes apart from each other. The proposed reservation ends too close to another reservation." in response.content.decode())

		tool.maximum_future_reservation_time = 90
		tool.minimum_time_between_reservations = None
		tool.save()
		data = self.get_reservation_data(start + timedelta(minutes=90), end + timedelta(minutes=90), tool)
		response = self.client.post(reverse('create_reservation'), data, follow=True)
		self.assertEquals(response.status_code, 200)
		self.assertTrue("You may only reserve up to 90 minutes of time on this tool, starting from the current time onward." in response.content.decode())

		first_of_the_day.delete()

	def test_create_reservation(self):
		start = datetime.now() + timedelta(hours=1)
		end = start + timedelta(hours=1)
		data = self.get_reservation_data(start, end, tool)

		login_as(self.client, consumer)
		response = self.client.post(reverse('create_reservation'), data, follow=True)
		self.assertEquals(response.status_code, 200)
		self.assertEquals(Reservation.objects.all().count(), 1)
		self.assertTrue(Reservation.objects.get(tool=tool))

	def test_create_reservation_multi_projects(self):
		start = datetime.now() + timedelta(hours=1)
		end = start + timedelta(hours=1)
		data = self.get_reservation_data(start, end, tool)

		login_as(self.client, consumer)

		second_project = Project.objects.create(name="project2", account=Account.objects.get(name='account1'))
		consumer.projects.add(second_project)

		# test creating reservation on a project the user is not associated with (trying to bill another project)
		not_my_project = Project.objects.create(name="not my project", account=Account.objects.get(name="account1"))
		data['project_id'] = not_my_project.id
		response = self.client.post(reverse('create_reservation'), data, follow=True)
		self.assertEquals(response.status_code, 200)
		self.assertEquals(Reservation.objects.all().count(), 0)
		self.assertTrue("Associate your reservation with a project." in response.content.decode())

		# test not sending project id
		data['project_id'] = ''
		response = self.client.post(reverse('create_reservation'), data, follow=True)
		self.assertEquals(response.status_code, 200)
		self.assertEquals(Reservation.objects.all().count(), 0)
		self.assertTrue("Associate your reservation with a project." in response.content.decode())

		data['project_id'] = second_project.id
		response = self.client.post(reverse('create_reservation'), data, follow=True)
		self.assertEquals(response.status_code, 200)
		self.assertEquals(Reservation.objects.all().count(), 1)
		self.assertTrue(Reservation.objects.get(tool=tool))

	def test_create_reservation_for_somebody_else(self):
		start = datetime.now() + timedelta(hours=1)
		end = start + timedelta(hours=1)
		data = self.get_reservation_data(start, end, tool)
		data['impersonate'] = consumer.id

		login_as(self.client, staff)
		response = self.client.post(reverse('create_reservation'), data, follow=True)
		self.assertEquals(response.status_code, 200)
		self.assertEquals(Reservation.objects.all().count(), 1)
		self.assertTrue(Reservation.objects.get(tool=tool))
		self.assertEquals(Reservation.objects.get(tool=tool).user, consumer)
		self.assertEquals(Reservation.objects.get(tool=tool).creator, staff)

	def test_resize_reservation(self):
		# create reservation
		start = datetime.now() + timedelta(hours=1)
		end = start + timedelta(hours=1)
		data = self.get_reservation_data(start, end, tool)

		login_as(self.client, consumer)
		response = self.client.post(reverse('create_reservation'), data, follow=True)
		self.assertEquals(response.status_code, 200)
		reservation = Reservation.objects.get(tool=tool)
		self.assertTrue(reservation.id)

		# test wrong delta
		response = self.client.post(reverse('resize_reservation'), {'delta': 'asd', 'id': reservation.id}, follow=True)
		self.assertEquals(response.status_code, 400)
		self.assertEquals(response.content.decode(), "Invalid delta")

		# test no reservation id
		response = self.client.post(reverse('resize_reservation'), {'delta': 10}, follow=True)
		self.assertEquals(response.status_code, 404)
		self.assertEquals(response.content.decode(), "The reservation that you wish to modify doesn't exist!")

		# test resize to less than original time
		response = self.client.post(reverse('resize_reservation'), {'delta': -60, 'id': reservation.id}, follow=True)
		self.assertEquals(response.status_code, 400)
		self.assertTrue('Reservation start time' in response.content.decode())
		self.assertTrue('must be before the end time' in response.content.decode())

		# test resize to end before now
		old_resa = Reservation.objects.create(tool=tool, start=datetime.now() - timedelta(hours=1), end=datetime.now() + timedelta(hours=1), creator=consumer, user=consumer, short_notice=False)
		response = self.client.post(reverse('resize_reservation'), {'delta': -65, 'id': old_resa.id}, follow=True)
		self.assertEquals(response.status_code, 400)
		self.assertTrue('Reservation start time' in response.content.decode())
		self.assertTrue('is earlier than the current time' in response.content.decode())
		old_resa.delete()

		# create a outage and try to resize reservation to overlap outage
		start_reservation = end + timedelta(hours=1)
		end_reservation = start_reservation + timedelta(hours=1)
		ScheduledOutage.objects.create(tool=tool, start=start_reservation, end=end_reservation, creator=staff)
		response = self.client.post(reverse('resize_reservation'), {'delta': 61, 'id': reservation.id}, follow=True)
		self.assertEquals(response.status_code, 400)
		self.assertEquals(response.content.decode(), "Your reservation coincides with a scheduled outage. Please choose a different time.")

		# test reduce reservation time by 10 min
		response = self.client.post(reverse('resize_reservation'), {'delta': -10, 'id': reservation.id}, follow=True)
		self.assertEquals(response.status_code, 200)
		old_reservation = Reservation.objects.get(pk=reservation.id)
		self.assertTrue(old_reservation.cancelled)
		self.assertEquals(old_reservation.cancelled_by, consumer)
		self.assertEquals(Reservation.objects.filter(tool=tool, cancelled=False).count(), 1)
		new_reservation = list(Reservation.objects.filter(tool=tool, cancelled=False))[0]
		self.assertEquals(new_reservation.end, old_reservation.end - timedelta(minutes=10))

		# test resize cancelled reservation
		reservation = Reservation.objects.get(pk=reservation.id)
		response = self.client.post(reverse('resize_reservation'), {'delta': 10, 'id': reservation.id}, follow=True)
		self.assertEquals(response.status_code, 400)
		self.assertTrue("This reservation has already been cancelled" in response.content.decode())

		# test increase reservation time by 10 min
		response = self.client.post(reverse('resize_reservation'), {'delta': 10, 'id': new_reservation.id}, follow=True)
		self.assertEquals(response.status_code, 200)
		old_reservation = Reservation.objects.get(pk=new_reservation.id)
		self.assertTrue(old_reservation.cancelled)
		self.assertEquals(old_reservation.cancelled_by, consumer)
		self.assertEquals(Reservation.objects.filter(tool=tool, cancelled=False).count(), 1)
		new_reservation = list(Reservation.objects.filter(tool=tool, cancelled=False))[0]
		self.assertEquals(new_reservation.end, old_reservation.end + timedelta(minutes=10))

	def test_move_reservation(self):
		# create reservation
		start = datetime.now() + timedelta(hours=1)
		end = start + timedelta(hours=1)
		data = self.get_reservation_data(start, end, tool)

		login_as(self.client, consumer)
		response = self.client.post(reverse('create_reservation'), data, follow=True)
		self.assertEquals(response.status_code, 200)
		reservation = Reservation.objects.get(tool=tool)
		self.assertTrue(reservation.id)

		# test wrong delta
		response = self.client.post(reverse('move_reservation'), {'delta': 'asd', 'id': reservation.id}, follow=True)
		self.assertEquals(response.status_code, 400)
		self.assertEquals(response.content.decode(), "Invalid delta")

		# test no reservation id
		response = self.client.post(reverse('move_reservation'), {'delta': 10}, follow=True)
		self.assertEquals(response.status_code, 404)
		self.assertEquals(response.content.decode(), "The reservation that you wish to modify doesn't exist!")

		# create a outage and try to move reservation to overlap reservation
		start_reservation = end + timedelta(hours=1)
		end_reservation = start_reservation + timedelta(hours=1)
		ScheduledOutage.objects.create(tool=tool, start=start_reservation, end=end_reservation, creator=staff)
		response = self.client.post(reverse('move_reservation'), {'delta': 61, 'id': reservation.id}, follow=True)
		self.assertEquals(response.status_code, 400)
		self.assertEquals(response.content.decode(), "Your reservation coincides with a scheduled outage. Please choose a different time.")

		# test move reservation 10 min earlier
		response = self.client.post(reverse('move_reservation'), {'delta': -10, 'id': reservation.id}, follow=True)
		self.assertEquals(response.status_code, 200)
		old_reservation = Reservation.objects.get(pk=reservation.id)
		self.assertTrue(old_reservation.cancelled)
		self.assertEquals(old_reservation.cancelled_by, consumer)
		self.assertEquals(Reservation.objects.filter(tool=tool, cancelled=False).count(), 1)
		new_reservation = list(Reservation.objects.filter(tool=tool, cancelled=False))[0]
		self.assertEquals(new_reservation.end, old_reservation.end - timedelta(minutes=10))
		self.assertEquals(new_reservation.start, old_reservation.start - timedelta(minutes=10))

		# test move cancelled reservation
		reservation = Reservation.objects.get(pk=reservation.id)
		response = self.client.post(reverse('move_reservation'), {'delta': 10, 'id': reservation.id}, follow=True)
		self.assertEquals(response.status_code, 400)
		self.assertTrue("This reservation has already been cancelled" in response.content.decode())

		# test move new reservation 10 min later
		response = self.client.post(reverse('move_reservation'), {'delta': 10, 'id': new_reservation.id}, follow=True)
		self.assertEquals(response.status_code, 200)
		old_reservation = Reservation.objects.get(pk=new_reservation.id)
		self.assertTrue(old_reservation.cancelled)
		self.assertEquals(old_reservation.cancelled_by, consumer)
		self.assertEquals(Reservation.objects.filter(tool=tool, cancelled=False).count(), 1)
		new_reservation = list(Reservation.objects.filter(tool=tool, cancelled=False))[0]
		self.assertEquals(new_reservation.end, old_reservation.end + timedelta(minutes=10))
		self.assertEquals(new_reservation.start, old_reservation.start + timedelta(minutes=10))

	def test_cancel_reservation(self):
		# create reservation
		start = datetime.now() + timedelta(hours=1)
		end = start + timedelta(hours=1)
		data = self.get_reservation_data(start, end, tool)

		login_as(self.client, consumer)
		response = self.client.post(reverse('create_reservation'), data, follow=True)
		self.assertEquals(response.status_code, 200)
		reservation = Reservation.objects.get(tool=tool)
		self.assertTrue(reservation.id)

		# get should fail
		response = self.client.get(reverse('cancel_reservation', kwargs={'reservation_id': 999}), {}, follow=True)
		self.assertEquals(response.status_code, 405)

		# test wrong id
		response = self.client.post(reverse('cancel_reservation', kwargs={'reservation_id': 999}), {}, follow=True)
		self.assertEquals(response.status_code, 404)

		# test non staff user trying to cancel reservation
		login_as_user(self.client)
		response = self.client.post(reverse('cancel_reservation', kwargs={'reservation_id': reservation.id}), {}, follow=True)
		self.assertEquals(response.status_code, 400)
		self.assertEquals("You must provide a reason when cancelling someone else's reservation.", response.content.decode())

		response = self.client.post(reverse('cancel_reservation', kwargs={'reservation_id': reservation.id}), {'reason':'reason'}, follow=True)
		self.assertEquals(response.status_code, 400)
		self.assertEquals("You may not cancel reservations that you do not own.", response.content.decode())

		login_as(self.client, consumer)

		# test cancel missed reservation
		missed_resa = Reservation.objects.create(tool=tool, start=start+timedelta(days=1), end=end+timedelta(days=1), user=consumer, creator=consumer, missed=True, short_notice=False)
		response = self.client.post(reverse('cancel_reservation', kwargs={'reservation_id': missed_resa.id}), {}, follow=True)
		self.assertEquals(response.status_code, 400)
		self.assertEquals(response.content.decode(), "This reservation was missed and cannot be modified.")

		# test cancel already ended reservation
		already_ended_resa = Reservation.objects.create(tool=tool, start=start - timedelta(days=1), end=end - timedelta(days=1), user=consumer, creator=consumer, missed=True, short_notice=False)
		response = self.client.post(reverse('cancel_reservation', kwargs={'reservation_id': already_ended_resa.id}), {}, follow=True)
		self.assertEquals(response.status_code, 400)
		self.assertEquals(response.content.decode(), "You may not cancel reservations that have already ended.")

		# cancel reservation
		response = self.client.post(reverse('cancel_reservation', kwargs={'reservation_id': reservation.id}), {}, follow=True)
		self.assertEquals(response.status_code, 200)
		self.assertTrue(Reservation.objects.get(pk=reservation.id).cancelled)
		self.assertEquals(Reservation.objects.get(pk=reservation.id).cancelled_by, consumer)

		# test cancel already cancelled reservation
		response = self.client.post(reverse('cancel_reservation', kwargs={'reservation_id': reservation.id}), {}, follow=True)
		self.assertEquals(response.status_code, 400)
		self.assertTrue("This reservation has already been cancelled by " in response.content.decode())

		# test staff cancelling somebody else's reservation
		other_resa = Reservation.objects.create(tool=tool, start=start - timedelta(days=1),	end=end - timedelta(days=1), user=consumer, creator=consumer, short_notice=False)
		login_as(self.client, staff)
		response = self.client.post(reverse('cancel_reservation', kwargs={'reservation_id': other_resa.id}), {}, follow=True)
		self.assertEquals(response.status_code, 400)
		self.assertEquals("You must provide a reason when cancelling someone else's reservation.", response.content.decode())

		response = self.client.post(reverse('cancel_reservation', kwargs={'reservation_id': other_resa.id}), {'reason': 'reason'}, follow=True)
		self.assertEquals(response.status_code, 200)
		self.assertTrue(Reservation.objects.get(pk=other_resa.id).cancelled)
		self.assertEquals(Reservation.objects.get(pk=other_resa.id).cancelled_by, staff)

	def test_reservation_details(self):
		# create reservation
		start = datetime.now() + timedelta(hours=1)
		end = start + timedelta(hours=1)
		data = self.get_reservation_data(start, end, tool)

		login_as(self.client, consumer)
		response = self.client.post(reverse('create_reservation'), data, follow=True)
		self.assertEquals(response.status_code, 200)
		reservation = Reservation.objects.get(tool=tool)
		self.assertTrue(reservation.id)

		# anybody that is logged in can see reservation details
		login_as_user(self.client)

		# post should fail
		response = self.client.post(reverse('reservation_details', kwargs={'reservation_id': 999}), {}, follow=True)
		self.assertEquals(response.status_code, 405)

		# test wrong id
		response = self.client.get(reverse('reservation_details', kwargs={'reservation_id': 999}), {}, follow=True)
		self.assertEquals(response.status_code, 404)

		response = self.client.get(reverse('reservation_details', kwargs={'reservation_id': reservation.id}), {}, follow=True)
		self.assertEquals(response.status_code, 200)

	def test_reservation_with_tool_configuration(self):
		# TODO: create those tests
		self.assertTrue(True)

	def test_reservation_policy_off(self):
		# TODO: create those tests
		self.assertTrue(True)



def disable_session_expiry_refresh(f):
	"""
	Security policy dictates that the user's session should time out after a set duration.
	The user's session is automatically refreshed (that is, the expiration date
	of their session is moved forward to 30 minutes after the request time) whenever
	the user performs an action. Pages such as the Calendar, Tool Control, and Status Dashboard
	all have polling AJAX requests to update information on the page. These regular polling
	requests should not refresh the session (because it does not indicate the user took
	an action). Place this decorator on any view that is regularly polled so that the
	user's session is not refreshed.
	"""
	f.disable_session_expiry_refresh = True
	return f



from _ssl import PROTOCOL_TLSv1_2, CERT_REQUIRED
from base64 import b64decode
from logging import exception, getLogger

from django.conf import settings
from django.contrib.auth import authenticate, login, REDIRECT_FIELD_NAME, logout
from django.contrib.auth.backends import RemoteUserBackend, ModelBackend
from django.http import HttpResponseRedirect
from django.shortcuts import render
from django.urls import reverse, resolve
from django.utils.decorators import method_decorator
from django.views.decorators.debug import sensitive_post_parameters
from django.views.decorators.http import require_http_methods, require_GET
from ldap3 import Tls, Server, Connection, AUTO_BIND_TLS_BEFORE_BIND, SIMPLE
from ldap3.core.exceptions import LDAPBindError, LDAPExceptionError

from NEMO.models import User
from NEMO.views.customization import get_media_file_contents

auth_logger = getLogger(__name__)

class RemoteUserAuthenticationBackend(RemoteUserBackend):
	""" The web server performs Kerberos authentication and passes the user name in via the REMOTE_USER environment variable. """
	create_unknown_user = False

	def clean_username(self, username):
		"""
		User names arrive in the form user@DOMAIN.NAME.
		This function chops off Kerberos realm information (i.e. the '@' and everything after).
		"""
		return username.partition('@')[0]


class NginxKerberosAuthorizationHeaderAuthenticationBackend(ModelBackend):
	""" The web server performs Kerberos authentication and passes the user name in via the HTTP_AUTHORIZATION header. """

	def authenticate(self, request, username=None, password=None, **keyword_arguments):
		# Perform any custom security checks below.
		# Returning None blocks the user's access.
		username = self.clean_username(request.META.get('HTTP_AUTHORIZATION', None))

		# The user must exist in the database
		try:
			user = User.objects.get(username=username)
		except User.DoesNotExist:
			auth_logger.warning(f"Username {username} attempted to authenticate with Kerberos via Nginx, but that username does not exist in the NEMO database. The user was denied access.")
			return None

		# The user must be marked active.
		if not user.is_active:
			auth_logger.warning(f"User {username} successfully authenticated with Kerberos via Nginx, but that user is marked inactive in the NEMO database. The user was denied access.")
			return None

		# All security checks passed so let the user in.
		auth_logger.debug(f"User {username} successfully authenticated with Kerberos via Nginx and was granted access to NEMO.")
		return user

	def clean_username(self, username):
		"""
		User names arrive encoded in base 64, similar to Basic authentication, but with a bogus password set (since .
		This function chops off Kerberos realm information (i.e. the '@' and everything after).
		"""
		if not username:
			return None
		pieces = username.split()
		if len(pieces) != 2:
			return None
		if pieces[0] != "Basic":
			return None
		return b64decode(pieces[1]).decode().partition(':')[0]


class LDAPAuthenticationBackend(ModelBackend):
	""" This class provides LDAP authentication against an LDAP or Active Directory server. """

	@method_decorator(sensitive_post_parameters('password'))
	def authenticate(self, request, username=None, password=None, **keyword_arguments):
		if not username or not password:
			return None

		# The user must exist in the database
		try:
			user = User.objects.get(username=username)
		except User.DoesNotExist:
			auth_logger.warning(f"Username {username} attempted to authenticate with LDAP, but that username does not exist in the NEMO database. The user was denied access.")
			return None

		# The user must be marked active.
		if not user.is_active:
			auth_logger.warning(f"User {username} successfully authenticated with LDAP, but that user is marked inactive in the NEMO database. The user was denied access.")
			return None

		for server in settings.LDAP_SERVERS:
			try:
				t = Tls(validate=CERT_REQUIRED, version=PROTOCOL_TLSv1_2, ca_certs_file=server['certificate'])
				s = Server(server['url'], port=636, use_ssl=True, tls=t)
				c = Connection(s, user='{}\\{}'.format(server['domain'], username), password=password, auto_bind=AUTO_BIND_TLS_BEFORE_BIND, authentication=SIMPLE)
				c.unbind()
				# At this point the user successfully authenticated to at least one LDAP server.
				return user
			except LDAPBindError as e:
				auth_logger.warning(f"User {username} attempted to authenticate with LDAP, but entered an incorrect password. The user was denied access.")
				pass  # When this error is caught it means the username and password were invalid against the LDAP server.
			except LDAPExceptionError as e:
				exception(e)

		# The user did not successfully authenticate to any of the LDAP servers.
		return None


@require_http_methods(['GET', 'POST'])
@sensitive_post_parameters('password')
def login_user(request):
	if 'NEMO.views.authentication.RemoteUserAuthenticationBackend' in settings.AUTHENTICATION_BACKENDS or 'NEMO.views.authentication.NginxKerberosAuthorizationHeaderAuthenticationBackend' in settings.AUTHENTICATION_BACKENDS:
		if request.user.is_authenticated:
			return HttpResponseRedirect(reverse('landing'))
		else:
			backends = [backend for backend in settings.AUTHENTICATION_BACKENDS if backend not in ['NEMO.views.authentication.RemoteUserAuthenticationBackend','NEMO.views.authentication.NginxKerberosAuthorizationHeaderAuthenticationBackend']]
			if len(backends) == 0:
				# there are no other authentication backends in the list, send error. Otherwise keep going
				return render(request, 'authorization_failed.html')

	dictionary = {
		'login_banner': get_media_file_contents('login_banner.html'),
		'user_name_or_password_incorrect': False,
	}
	if request.method == 'GET':
		return render(request, 'login.html', dictionary)
	username = request.POST.get('username', '')
	password = request.POST.get('password', '')
	user = authenticate(request, username=username, password=password)
	if user:
		login(request, user)
		try:
			next_page = request.GET[REDIRECT_FIELD_NAME]
			resolve(next_page)  # Make sure the next page is a legitimate URL for NEMO
		except:
			next_page = reverse('landing')
		return HttpResponseRedirect(next_page)
	dictionary['user_name_or_password_incorrect'] = True
	return render(request, 'login.html', dictionary)


@require_GET
def logout_user(request):
	logout(request)
	return HttpResponseRedirect(reverse('landing'))



from rest_framework.fields import FloatField, CharField, IntegerField, DateTimeField, ChoiceField
from rest_framework.serializers import Serializer, ModelSerializer

from NEMO.models import User, Project, Account, Reservation, AreaAccessRecord, UsageEvent, Task, TaskHistory, ScheduledOutage, Tool


class UserSerializer(ModelSerializer):
	class Meta:
		model = User
		fields = ('id', 'first_name', 'last_name', 'username', 'email', 'date_joined', 'badge_number', 'is_active')


class ProjectSerializer(ModelSerializer):
	class Meta:
		model = Project
		fields = ('id', 'name', 'application_identifier', 'active')


class AccountSerializer(ModelSerializer):
	class Meta:
		model = Account
		fields = ('id', 'name', 'active')


class ToolSerializer(ModelSerializer):
	class Meta:
		model = Tool
		fields = '__all__'


class ReservationSerializer(ModelSerializer):
	class Meta:
		model = Reservation
		fields = '__all__'


class UsageEventSerializer(ModelSerializer):
	class Meta:
		model = UsageEvent
		fields = '__all__'


class AreaAccessRecordSerializer(ModelSerializer):
	class Meta:
		model = AreaAccessRecord
		fields = '__all__'


class TaskHistorySerializer(ModelSerializer):
	class Meta:
		model = TaskHistory
		fields = '__all__'


class TaskSerializer(ModelSerializer):
	history = TaskHistorySerializer(many=True, read_only=True)

	class Meta:
		model = Task
		fields = '__all__'


class ScheduledOutageSerializer(ModelSerializer):
	class Meta:
		model = ScheduledOutage
		fields = '__all__'


class BillableItemSerializer(Serializer):
	type = ChoiceField(['missed_reservation', 'tool_usage', 'area_access', 'consumable'])
	details = CharField(max_length=500, read_only=True)
	account = CharField(max_length=200, read_only=True)
	account_id = IntegerField(read_only=True)
	project = CharField(max_length=200, read_only=True)
	project_id = IntegerField(read_only=True)
	application = CharField(max_length=200, read_only=True)
	username = CharField(max_length=200, read_only=True)
	user_id = IntegerField(read_only=True)
	start = DateTimeField(read_only=True)
	end = DateTimeField(read_only=True)
	quantity = FloatField(read_only=True)

	def update(self, instance, validated_data):
		pass

	def create(self, validated_data):
		pass

	class Meta:
		fields = '__all__'



from django.contrib.auth.decorators import login_required
from django.shortcuts import render
from django.template import Template, Context
from django.views.decorators.http import require_http_methods

from NEMO.utilities import parse_parameter_string, send_mail
from NEMO.views.constants import FEEDBACK_MAXIMUM_LENGTH
from NEMO.views.customization import get_customization, get_media_file_contents


@login_required
@require_http_methods(['GET', 'POST'])
def feedback(request):
	recipient = get_customization('feedback_email_address')
	email_contents = get_media_file_contents('feedback_email.html')
	if not recipient or not email_contents:
		return render(request, 'feedback.html', {'customization_required': True})

	if request.method == 'GET':
		return render(request, 'feedback.html')
	contents = parse_parameter_string(request.POST, 'feedback', FEEDBACK_MAXIMUM_LENGTH)
	if contents == '':
		return render(request, 'feedback.html')
	dictionary = {
		'contents': contents,
		'user': request.user,
	}

	email = Template(email_contents).render(Context(dictionary))
	send_mail('Feedback from ' + str(request.user), email, request.user.email, [recipient])
	dictionary = {
		'title': 'Feedback',
		'heading': 'Thanks for your feedback!',
		'content': 'Your feedback has been sent to the NanoFab staff. We will follow up with you as soon as we can.',
	}
	return render(request, 'acknowledgement.html', dictionary)



from django.contrib.admin.views.decorators import staff_member_required
from django.http import Http404
from django.shortcuts import render, get_object_or_404, redirect
from django.views.decorators.http import require_GET, require_http_methods, require_POST

from NEMO.forms import ScheduledOutageForm
from NEMO.models import Resource, UsageEvent, Tool, ScheduledOutage, ScheduledOutageCategory


@staff_member_required(login_url=None)
@require_GET
def resources(request):
	return render(request, 'resources/resources.html', {'resources': Resource.objects.all().order_by('category', 'name')})


@staff_member_required(login_url=None)
@require_http_methods(['GET', 'POST'])
def modify_resource(request, resource_id):
	resource = get_object_or_404(Resource, id=resource_id)
	dictionary = {'resource': resource}
	if request.method == 'GET':
		in_use = set(map(lambda t: t.tool.tool_or_parent_id(), UsageEvent.objects.filter(end=None)))
		fully_dependent_tools = set(map(lambda f: f.id, resource.fully_dependent_tools.all()))
		fully_dependent_tools_in_use = Tool.objects.in_bulk(list(in_use & fully_dependent_tools)).values()
		dictionary['fully_dependent_tools_in_use'] = fully_dependent_tools_in_use
		return render(request, 'resources/modify_resource.html', dictionary)
	elif request.method == 'POST':
		status = request.POST.get('status')
		if status == 'unavailable':
			resource.available = False
			reason = request.POST.get('reason')
			if not reason:
				dictionary['error'] = 'You must explain why the resource is unavailable.'
				return render(request, 'resources/modify_resource.html', dictionary)
			resource.restriction_message = reason
			resource.save()
		elif status == 'available':
			resource.available = True
			resource.save()
		else:
			dictionary['error'] = 'The server received an invalid resource modification request.'
			return render(request, 'resources/modify_resource.html', dictionary)
		return redirect('resources')


@staff_member_required(login_url=None)
@require_http_methods(['GET', 'POST'])
def schedule_outage(request):
	outage_id = request.GET.get('outage_id') or request.POST.get('outage_id')
	try:
		outage = ScheduledOutage.objects.get(id=outage_id)
	except:
		outage = None
	if request.method == 'GET':
		form = ScheduledOutageForm(instance=outage)
	elif request.method == 'POST':
		form = ScheduledOutageForm(data=request.POST, instance=outage)
		if form.is_valid():
			outage = form.save(commit=False)
			outage.creator = request.user
			outage.title = f"{outage.resource.name} scheduled outage"
			outage.save()
			form = ScheduledOutageForm()
	else:
		form = ScheduledOutageForm()
	dictionary = {
		'form': form,
		'editing': True if form.instance.id else False,
		'resources': Resource.objects.all().prefetch_related('category').order_by('category__name', 'name'),
		'outages': ScheduledOutage.objects.filter(resource__isnull=False),
		'outage_categories': ScheduledOutageCategory.objects.all(),
	}
	return render(request, 'resources/scheduled_outage.html', dictionary)


@staff_member_required(login_url=None)
@require_POST
def delete_scheduled_outage(request, outage_id):
	try:
		ScheduledOutage.objects.filter(id=outage_id).delete()
	except Http404:
		pass
	return redirect(request.META.get('HTTP_REFERER', 'landing'))



from django.test import TestCase
from django.urls import reverse

from NEMO.models import User, Tool
from NEMO.tests.test_utilities import login_as_user, test_response_is_login_page


class CalendarTestCase(TestCase):
	tool = None
	owner = None

	def setUp(self):
		global tool, owner
		owner = User.objects.create(username='mctest', first_name='Testy', last_name='McTester')
		tool = Tool.objects.create(name='test_tool', primary_owner=owner, _category='Imaging')

	def test_calendar_urls(self):
		# if not logged in, it should redirect to login
		response = self.client.get(reverse('calendar'), follow=True)
		test_response_is_login_page(self, response)

		login_as_user(self.client)
		response = self.client.get(reverse('calendar'), follow=True)
		self.assertEquals(response.status_code, 200)

		response = self.client.get(reverse('calendar', kwargs={'tool_id': tool.id}), follow=True)
		self.assertEquals(response.status_code, 200)



from django.contrib.auth.decorators import login_required
from django.db.models import Q
from django.http import HttpResponse
from django.shortcuts import render
from django.utils import timezone
from django.views.decorators.http import require_GET

from NEMO.decorators import disable_session_expiry_refresh
from NEMO.models import Area, AreaAccessRecord, Resource, ScheduledOutage, Task, Tool, UsageEvent


@login_required
@require_GET
@disable_session_expiry_refresh
def status_dashboard(request, tab=None):
	"""
	Present a web page to allow users to view the status and usage of all tools.
	"""
	interest = request.GET.get('interest')
	if interest is None:
		dictionary = {
			'tab': tab if tab else "occupancy",
			'tool_summary': create_tool_summary(),
			'nanofab_occupants': AreaAccessRecord.objects.filter(end=None, staff_charge=None).prefetch_related('customer', 'project', 'area'),
		}
		return render(request, 'status_dashboard/status_dashboard.html', dictionary)
	elif interest == "tools":
		dictionary = {
			'tool_summary': create_tool_summary(),
		}
		return render(request, 'status_dashboard/tools.html', dictionary)
	elif interest == "occupancy":
		dictionary = {
			'nanofab_occupants': AreaAccessRecord.objects.filter(end=None, staff_charge=None).prefetch_related('customer', 'project', 'area'),
		}
		return render(request, 'status_dashboard/occupancy.html', dictionary)


def create_tool_summary():
	tools = Tool.objects.filter(visible=True)
	tasks = Task.objects.filter(cancelled=False, resolved=False, tool__visible=True).prefetch_related('tool')
	unavailable_resources = Resource.objects.filter(available=False).prefetch_related('fully_dependent_tools', 'partially_dependent_tools')
	# also check for visibility on the parent if there is one (alternate tool are hidden)
	usage_events = UsageEvent.objects.filter(Q(end=None, tool__visible=True)|Q(end=None, tool__parent_tool__visible=True)).prefetch_related('operator', 'user', 'tool')
	scheduled_outages = ScheduledOutage.objects.filter(start__lte=timezone.now(), end__gt=timezone.now())
	tool_summary = merge(tools, tasks, unavailable_resources, usage_events, scheduled_outages)
	tool_summary = list(tool_summary.values())
	tool_summary.sort(key=lambda x: x['name'])
	return tool_summary


def merge(tools, tasks, unavailable_resources, usage_events, scheduled_outages):
	result = {}
	tools_with_delayed_logoff_in_effect = [x.tool.tool_or_parent_id() for x in UsageEvent.objects.filter(end__gt=timezone.now())]
	for tool in tools:
		result[tool.tool_or_parent_id()] = {
			'name': tool.name_or_child_in_use_name(),
			'id': tool.id,
			'user': '',
			'operator': '',
			'in_use': False,
			'in_use_since': '',
			'delayed_logoff_in_progress': tool.tool_or_parent_id() in tools_with_delayed_logoff_in_effect,
			'problematic': False,
			'operational': tool.operational,
			'required_resource_is_unavailable': False,
			'nonrequired_resource_is_unavailable': False,
			'scheduled_outage': False,
		}
	for task in tasks:
		result[task.tool.id]['problematic'] = True
	for event in usage_events:
		result[event.tool.tool_or_parent_id()]['operator'] = str(event.operator)
		result[event.tool.tool_or_parent_id()]['user'] = str(event.operator)
		if event.user != event.operator:
			result[event.tool.tool_or_parent_id()]['user'] += " on behalf of " + str(event.user)
		result[event.tool.tool_or_parent_id()]['in_use'] = True
		result[event.tool.tool_or_parent_id()]['in_use_since'] = event.start
	for resource in unavailable_resources:
		for tool in resource.fully_dependent_tools.filter(visible=True):
			result[tool.id]['required_resource_is_unavailable'] = True
		for tool in resource.partially_dependent_tools.filter(visible=True):
			result[tool.id]['nonrequired_resource_is_unavailable'] = True
	for outage in scheduled_outages:
		if outage.tool_id and outage.tool.visible:
			result[outage.tool.id]['scheduled_outage'] = True
		elif outage.resource_id:
			for t in outage.resource.fully_dependent_tools.filter(visible=True):
				result[t.id]['scheduled_outage'] = True
	return result


@login_required
@require_GET
@disable_session_expiry_refresh
def occupancy(request):
	area = request.GET.get('occupancy')
	if area is None or not Area.objects.filter(name=area).exists():
		return HttpResponse()
	dictionary = {
		'area': area,
		'occupants': AreaAccessRecord.objects.filter(area__name=area, end=None, staff_charge=None).prefetch_related('customer'),
	}
	return render(request, 'occupancy.html', dictionary)



import struct
from socket import socket
from unittest import mock

from django.conf import settings
from django.test import TestCase

# This method will be used by the mock to replace socket.send
# In the stanford interlock case, it will return a list of byte
from NEMO.models import Tool, Interlock, InterlockCard, InterlockCardCategory, User

server1 = 'server1.nist.gov'
server2 = 'https://server2.nist.gov'
port1 = 80
port2 = 8080

def mocked_socket_send(*args, **kwargs):
	class MockSocket(socket):

		def __init__(self):
			super().__init__()
			self.response = None
			self.address = None

		def send(self, data: bytes, flags: int = ...) -> int:
			schema = struct.Struct('!20siiiiiiiiibbbbb18s')
			message_to_be_sent = schema.unpack(data)
			card_number = message_to_be_sent[2]
			even_port = message_to_be_sent[3]
			odd_port = message_to_be_sent[4]
			channel = message_to_be_sent[5]
			command_type = message_to_be_sent[7]
			command_result = 1 # success

			# server2 port 2 channel 2 will send a command failed back
			if self.address[0] == server2 and self.address[1] == port2 and channel == 2:
				command_result = 0 # fail

			# created the response to be sent later
			response_schema = struct.Struct('!iiiiiiiiibbbbb')
			self.response = response_schema.pack(
				1,  # Instruction count
				card_number,
				even_port,
				odd_port,
				channel,
				command_result,  # Command return value.
				0,  # Instruction type
				0,  # Instruction
				0,  # Delay
				0,  # SD overload
				0,  # RD overload
				0,  # ADC done
				0,  # Busy
				0,  # Instruction return value
			)
			pass

		def connect(self, address) -> None:
			self.address = address
			pass

		def recv(self, bufsize: int, flags: int = ...) -> bytes:
			return self.response

	return MockSocket()

class StanfordInterlockTestCase(TestCase):
	tool: Tool = None
	wrong_response_interlock: Interlock = None
	bad_interlock: Interlock = None

	def setUp(self):
		global tool, wrong_response_interlock, bad_interlock
		# enable interlock functionality
		settings.__setattr__('INTERLOCKS_ENABLED', True)
		even_port = 124
		odd_port = 125
		interlock_card_category = InterlockCardCategory.objects.get(key='stanford')
		interlock_card = InterlockCard.objects.create(server=server1, port=port1, number=1, even_port=even_port, odd_port=odd_port, category=interlock_card_category)
		interlock_card2 = InterlockCard.objects.create(server=server2, port=port2, number=1, even_port=even_port, odd_port=odd_port, category=interlock_card_category)
		interlock = Interlock.objects.create(card=interlock_card, channel=1)
		wrong_response_interlock = Interlock.objects.create(card=interlock_card2, channel=2)
		bad_interlock = Interlock.objects.create(card=interlock_card2, channel=3)
		owner = User.objects.create(username='mctest', first_name='Testy', last_name='McTester')
		tool = Tool.objects.create(name='test_tool', primary_owner=owner, interlock=interlock)

	@mock.patch('NEMO.interlocks.socket.socket', side_effect=mocked_socket_send)
	def test_all_good(self,  mock_args):
		self.assertTrue(tool.interlock.unlock())
		self.assertEquals(tool.interlock.state, Interlock.State.UNLOCKED)
		self.assertTrue(tool.interlock.lock())
		self.assertEquals(tool.interlock.state, Interlock.State.LOCKED)

	@mock.patch('NEMO.interlocks.socket.socket', side_effect=mocked_socket_send)
	def test_server2_command_fail(self, mock_args):
		self.assertFalse(wrong_response_interlock.unlock())
		self.assertEquals(wrong_response_interlock.state, Interlock.State.UNKNOWN)
		self.assertTrue('failed' in wrong_response_interlock.most_recent_reply)
		self.assertTrue('command return value = 0' in wrong_response_interlock.most_recent_reply)
		self.assertFalse(wrong_response_interlock.lock())
		self.assertEquals(wrong_response_interlock.state, Interlock.State.UNKNOWN)
		self.assertTrue('failed' in wrong_response_interlock.most_recent_reply)
		self.assertTrue('command return value = 0' in wrong_response_interlock.most_recent_reply)




import io
from collections import Iterable
from datetime import timedelta, datetime
from http import HTTPStatus
from re import match
from typing import Union

from dateutil import rrule
from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.auth.decorators import login_required, permission_required
from django.db.models import Q
from django.http import HttpResponseBadRequest, HttpResponse, HttpResponseNotFound
from django.shortcuts import render, get_object_or_404, redirect
from django.template import Template, Context
from django.utils import timezone
from django.views.decorators.http import require_GET, require_POST

from NEMO.decorators import disable_session_expiry_refresh
from NEMO.models import Tool, Reservation, Configuration, UsageEvent, AreaAccessRecord, StaffCharge, User, Project, ScheduledOutage, ScheduledOutageCategory
from NEMO.utilities import bootstrap_primary_color, extract_times, extract_dates, format_datetime, parse_parameter_string, send_mail, create_email_attachment, localize
from NEMO.views.constants import ADDITIONAL_INFORMATION_MAXIMUM_LENGTH
from NEMO.views.customization import get_customization, get_media_file_contents
from NEMO.views.policy import check_policy_to_save_reservation, check_policy_to_cancel_reservation, check_policy_to_create_outage
from NEMO.widgets.tool_tree import ToolTree


recurrence_frequency_display = {
	'DAILY': 'Day(s)',
	'DAILY_WEEKDAYS':'Week Day(s)',
	'DAILY WEEKENDS':'Weekend Day(s)',
	'WEEKLY':'Week(s)',
	'MONTHLY':'Month(s)',
}

recurrence_frequencies = {
	'DAILY': rrule.DAILY,
	'DAILY_WEEKDAYS': rrule.DAILY,
	'DAILY WEEKENDS': rrule.DAILY,
	'WEEKLY': rrule.WEEKLY,
	'MONTHLY': rrule.MONTHLY,
}

@login_required
@require_GET
def calendar(request, tool_id=None):
	""" Present the calendar view to the user. """

	if request.device == 'mobile':
		if tool_id:
			return redirect('view_calendar', tool_id)
		else:
			return redirect('choose_tool', 'view_calendar')

	tools = Tool.objects.filter(visible=True).order_by('_category', 'name')
	rendered_tool_tree_html = ToolTree().render(None, {'tools': tools, 'user': request.user})
	dictionary = {
		'rendered_tool_tree_html': rendered_tool_tree_html,
		'tools': tools,
		'auto_select_tool': tool_id,
	}
	if request.user.is_staff:
		dictionary['users'] = User.objects.all()
	return render(request, 'calendar/calendar.html', dictionary)


@login_required
@require_GET
@disable_session_expiry_refresh
def event_feed(request):
	""" Get all reservations for a specific time-window. Optionally: filter by tool or user name. """
	try:
		start, end = extract_dates(request.GET)
	except Exception as e:
		return HttpResponseBadRequest('Invalid start or end time. ' + str(e))

	# We don't want to let someone hammer the database with phony calendar feed lookups.
	# Block any requests that have a duration of more than 8 weeks. The FullCalendar
	# should only ever request 6 weeks of data at a time (at most).
	if end - start > timedelta(weeks=8):
		return HttpResponseBadRequest("Calendar feed request has too long a duration: " + str(end - start))

	event_type = request.GET.get('event_type')

	if event_type == 'reservations':
		return reservation_event_feed(request, start, end)
	elif event_type == 'nanofab usage':
		return usage_event_feed(request, start, end)
	# Only staff may request a specific user's history...
	elif event_type == 'specific user' and request.user.is_staff:
		user = get_object_or_404(User, id=request.GET.get('user'))
		return specific_user_feed(request, user, start, end)
	else:
		return HttpResponseBadRequest('Invalid event type or operation not authorized.')


def reservation_event_feed(request, start, end):
	events = Reservation.objects.filter(cancelled=False, missed=False, shortened=False)
	outages = None
	# Exclude events for which the following is true:
	# The event starts and ends before the time-window, and...
	# The event starts and ends after the time-window.
	events = events.exclude(start__lt=start, end__lt=start)
	events = events.exclude(start__gt=end, end__gt=end)

	# Filter events that only have to do with the relevant tool.
	tool = request.GET.get('tool_id')
	if tool:
		events = events.filter(tool__id=tool)

		outages = ScheduledOutage.objects.filter(Q(tool=tool) | Q(resource__fully_dependent_tools__in=[tool]))
		outages = outages.exclude(start__lt=start, end__lt=start)
		outages = outages.exclude(start__gt=end, end__gt=end)

	# Filter events that only have to do with the current user.
	personal_schedule = request.GET.get('personal_schedule')
	if personal_schedule:
		events = events.filter(user=request.user)

	dictionary = {
		'events': events,
		'outages': outages,
		'personal_schedule': personal_schedule,
	}
	return render(request, 'calendar/reservation_event_feed.html', dictionary)


def usage_event_feed(request, start, end):
	usage_events = UsageEvent.objects
	# Exclude events for which the following is true:
	# The event starts and ends before the time-window, and...
	# The event starts and ends after the time-window.
	usage_events = usage_events.exclude(start__lt=start, end__lt=start)
	usage_events = usage_events.exclude(start__gt=end, end__gt=end)

	# Filter events that only have to do with the relevant tool.
	tool_id = request.GET.get('tool_id')
	if tool_id:
		usage_events = usage_events.filter(tool__id__in=Tool.objects.get(pk=tool_id).get_family_tool_ids())

	area_access_events = None
	# Filter events that only have to do with the current user.
	personal_schedule = request.GET.get('personal_schedule')
	if personal_schedule:
		usage_events = usage_events.filter(user=request.user)
		# Display area access along side tool usage when 'personal schedule' is selected.
		area_access_events = AreaAccessRecord.objects.filter(customer__id=request.user.id)
		area_access_events = area_access_events.exclude(start__lt=start, end__lt=start)
		area_access_events = area_access_events.exclude(start__gt=end, end__gt=end)

	missed_reservations = None
	if personal_schedule:
		missed_reservations = Reservation.objects.filter(missed=True, user=request.user)
	elif tool_id:
		missed_reservations = Reservation.objects.filter(missed=True, tool=tool_id)
	if missed_reservations:
		missed_reservations = missed_reservations.exclude(start__lt=start, end__lt=start)
		missed_reservations = missed_reservations.exclude(start__gt=end, end__gt=end)

	dictionary = {
		'usage_events': usage_events,
		'area_access_events': area_access_events,
		'personal_schedule': personal_schedule,
		'missed_reservations': missed_reservations,
	}
	return render(request, 'calendar/usage_event_feed.html', dictionary)


def specific_user_feed(request, user, start, end):
	# Find all tool usage events for a user.
	# Exclude events for which the following is true:
	# The event starts and ends before the time-window, and...
	# The event starts and ends after the time-window.
	usage_events = UsageEvent.objects.filter(user=user)
	usage_events = usage_events.exclude(start__lt=start, end__lt=start)
	usage_events = usage_events.exclude(start__gt=end, end__gt=end)

	# Find all area access events for a user.
	area_access_events = AreaAccessRecord.objects.filter(customer=user)
	area_access_events = area_access_events.exclude(start__lt=start, end__lt=start)
	area_access_events = area_access_events.exclude(start__gt=end, end__gt=end)

	# Find all reservations for the user that were not missed or cancelled.
	reservations = Reservation.objects.filter(user=user, missed=False, cancelled=False, shortened=False)
	reservations = reservations.exclude(start__lt=start, end__lt=start)
	reservations = reservations.exclude(start__gt=end, end__gt=end)

	# Find all missed reservations for the user.
	missed_reservations = Reservation.objects.filter(user=user, missed=True)
	missed_reservations = missed_reservations.exclude(start__lt=start, end__lt=start)
	missed_reservations = missed_reservations.exclude(start__gt=end, end__gt=end)

	dictionary = {
		'usage_events': usage_events,
		'area_access_events': area_access_events,
		'reservations': reservations,
		'missed_reservations': missed_reservations,
	}
	return render(request, 'calendar/specific_user_feed.html', dictionary)


@login_required
@require_POST
def create_reservation(request):
	""" Create a reservation for a user. """
	try:
		start, end = extract_times(request.POST)
	except Exception as e:
		return HttpResponseBadRequest(str(e))
	tool = get_object_or_404(Tool, name=request.POST.get('tool_name'))
	explicit_policy_override = False
	if request.user.is_staff:
		try:
			user = User.objects.get(id=request.POST['impersonate'])
		except:
			user = request.user
		try:
			explicit_policy_override = request.POST['explicit_policy_override'] == 'true'
		except:
			pass
	else:
		user = request.user
	# Create the new reservation:
	new_reservation = Reservation()
	new_reservation.user = user
	new_reservation.creator = request.user
	new_reservation.tool = tool
	new_reservation.start = start
	new_reservation.end = end
	new_reservation.short_notice = determine_insufficient_notice(tool, start)
	policy_problems, overridable = check_policy_to_save_reservation(None, new_reservation, user, explicit_policy_override)

	# If there was a problem in saving the reservation then return the error...
	if policy_problems:
		return render(request, 'calendar/policy_dialog.html', {'policy_problems': policy_problems, 'overridable': overridable and request.user.is_staff})

	# All policy checks have passed.

	# If the user only has one project then associate it with the reservation.
	# Otherwise, present a dialog box for the user to choose which project to associate.
	active_projects = user.active_projects()
	if len(active_projects) == 1:
		new_reservation.project = active_projects[0]
	else:
		try:
			new_reservation.project = Project.objects.get(id=request.POST['project_id'])
		except:
			return render(request, 'calendar/project_choice.html', {'active_projects': active_projects})

	# Make sure the user is actually enrolled on the project. We wouldn't want someone
	# forging a request to reserve against a project they don't belong to.
	if new_reservation.project not in new_reservation.user.active_projects():
		return render(request, 'calendar/project_choice.html', {'active_projects': active_projects})

	configured = (request.POST.get('configured') == "true")
	# If a reservation is requested and the tool does not require configuration...
	if not tool.is_configurable():
		new_reservation.save_and_notify()
		return HttpResponse()

	# If a reservation is requested and the tool requires configuration that has not been submitted...
	elif tool.is_configurable() and not configured:
		configuration_information = tool.get_configuration_information(user=user, start=start)
		return render(request, 'calendar/configuration.html', configuration_information)

	# If a reservation is requested and configuration information is present also...
	elif tool.is_configurable() and configured:
		new_reservation.additional_information, new_reservation.self_configuration = extract_configuration(request)
		# Reservation can't be short notice if the user is configuring the tool themselves.
		if new_reservation.self_configuration:
			new_reservation.short_notice = False
		new_reservation.save_and_notify()
		return HttpResponse()

	return HttpResponseBadRequest("Reservation creation failed because invalid parameters were sent to the server.")


def extract_configuration(request):
	cleaned_configuration = []
	for key, value in request.POST.items():
		entry = parse_configuration_entry(key, value)
		if entry:
			cleaned_configuration.append(entry)
	# Sort by configuration display priority and join the results:
	result = ''
	for config in sorted(cleaned_configuration):
		result += config[1] + '\n'
	if 'additional_information' in request.POST:
		result += request.POST['additional_information'][:ADDITIONAL_INFORMATION_MAXIMUM_LENGTH].strip()
	self_configuration = True if request.POST.get('self_configuration') == 'on' else False
	return result, self_configuration


def parse_configuration_entry(key, value):
	if value == "" or not match("^configuration_[0-9]+__slot_[0-9]+__display_priority_[0-9]+$", key):
		return None
	config_id, slot, display_priority = [int(s) for s in key.split('_') if s.isdigit()]
	configuration = Configuration.objects.get(pk=config_id)
	available_setting = configuration.get_available_setting(value)
	if len(configuration.current_settings_as_list()) == 1:
		return display_priority, configuration.name + " needs to be set to " + available_setting + "."
	else:
		return display_priority, configuration.configurable_item_name + " #" + str(slot + 1) + " needs to be set to " + available_setting + "."


@staff_member_required(login_url=None)
@require_POST
def create_outage(request):
	""" Create an outage. """
	try:
		start, end = extract_times(request.POST)
	except Exception as e:
		return HttpResponseBadRequest(str(e))
	tool = get_object_or_404(Tool, name=request.POST.get('tool_name'))
	# Create the new reservation:
	outage = ScheduledOutage()
	outage.creator = request.user
	outage.category = request.POST.get('category', '')[:200]
	outage.tool = tool
	outage.start = start
	outage.end = end

	# If there is a policy problem for the outage then return the error...
	policy_problem = check_policy_to_create_outage(outage)
	if policy_problem:
		return HttpResponseBadRequest(policy_problem)

	# Make sure there is at least an outage title
	if not request.POST.get('title'):
		dictionary = {
			'categories': ScheduledOutageCategory.objects.all(),
			'recurrence_intervals': recurrence_frequency_display,
			'recurrence_date_start': start.date(),
		}
		return render(request, 'calendar/scheduled_outage_information.html', dictionary)

	outage.title = request.POST['title']
	outage.details = request.POST.get('details', '')

	if request.POST.get('recurring_outage') == 'on':
		# we have to remove tz before creating rules otherwise 8am would become 7am after DST change for example.
		start_no_tz = outage.start.replace(tzinfo=None)
		end_no_tz = outage.end.replace(tzinfo=None)

		submitted_frequency = request.POST.get('recurrence_frequency')
		submitted_date_until = request.POST.get('recurrence_until', None)
		date_until = end.replace(hour=0, minute=0, second=0)
		if submitted_date_until:
			date_until = localize(datetime.strptime(submitted_date_until, '%m/%d/%Y'))
		date_until += timedelta(days=1, seconds=-1) # set at the end of the day
		by_week_day = None
		if submitted_frequency == 'DAILY_WEEKDAYS':
			by_week_day = (rrule.MO, rrule.TU, rrule.WE, rrule.TH, rrule.FR)
		elif submitted_frequency == 'DAILY_WEEKENDS':
			by_week_day = (rrule.SA, rrule.SU)
		frequency = recurrence_frequencies.get(submitted_frequency, rrule.DAILY)
		rules: Iterable[datetime] = rrule.rrule(dtstart=start, freq=frequency, interval=int(request.POST.get('recurrence_interval',1)), until=date_until, byweekday=by_week_day)
		for rule in list(rules):
			recurring_outage = ScheduledOutage()
			recurring_outage.creator = outage.creator
			recurring_outage.category = outage.category
			recurring_outage.tool = outage.tool
			recurring_outage.title = outage.title
			recurring_outage.details = outage.details
			recurring_outage.start = localize(start_no_tz.replace(year=rule.year, month=rule.month, day=rule.day))
			recurring_outage.end = localize(end_no_tz.replace(year=rule.year, month=rule.month, day=rule.day))
			recurring_outage.save()
	else:
		outage.save()

	return HttpResponse()


@login_required
@require_POST
def resize_reservation(request):
	""" Resize a reservation for a user. """
	try:
		delta = timedelta(minutes=int(request.POST['delta']))
	except:
		return HttpResponseBadRequest('Invalid delta')
	return modify_reservation(request, None, delta)


@staff_member_required(login_url=None)
@require_POST
def resize_outage(request):
	""" Resize an outage """
	try:
		delta = timedelta(minutes=int(request.POST['delta']))
	except:
		return HttpResponseBadRequest('Invalid delta')
	return modify_outage(request, None, delta)


@login_required
@require_POST
def move_reservation(request):
	""" Move a reservation for a user. """
	try:
		delta = timedelta(minutes=int(request.POST['delta']))
	except:
		return HttpResponseBadRequest('Invalid delta')
	return modify_reservation(request, delta, delta)


@staff_member_required(login_url=None)
@require_POST
def move_outage(request):
	""" Move a reservation for a user. """
	try:
		delta = timedelta(minutes=int(request.POST['delta']))
	except:
		return HttpResponseBadRequest('Invalid delta')
	return modify_outage(request, delta, delta)


def modify_reservation(request, start_delta, end_delta):
	"""
	Cancel the user's old reservation and create a new one. Reservations are cancelled and recreated so that
	reservation abuse can be tracked if necessary. This function should be called by other views and should
	not be tied directly to a URL.
	"""
	try:
		reservation_to_cancel = Reservation.objects.get(pk=request.POST.get('id'))
	except Reservation.DoesNotExist:
		return HttpResponseNotFound("The reservation that you wish to modify doesn't exist!")
	response = check_policy_to_cancel_reservation(reservation_to_cancel, request.user)
	# Do not move the reservation if the user was not authorized to cancel it.
	if response.status_code != HTTPStatus.OK:
		return response
	# Record the current time so that the timestamp of the cancelled reservation and the new reservation match exactly.
	now = timezone.now()
	# Cancel the user's original reservation.
	reservation_to_cancel.cancelled = True
	reservation_to_cancel.cancellation_time = now
	reservation_to_cancel.cancelled_by = request.user
	# Create a new reservation for the user.
	new_reservation = Reservation()
	new_reservation.title = reservation_to_cancel.title
	new_reservation.creator = request.user
	new_reservation.additional_information = reservation_to_cancel.additional_information
	# A change in start time will only be provided if the reservation is being moved.
	new_reservation.start = reservation_to_cancel.start
	new_reservation.self_configuration = reservation_to_cancel.self_configuration
	new_reservation.short_notice = False
	if start_delta:
		new_reservation.start += start_delta
	if new_reservation.self_configuration:
		# Reservation can't be short notice since the user is configuring the tool themselves.
		new_reservation.short_notice = False
	else:
		new_reservation.short_notice = determine_insufficient_notice(reservation_to_cancel.tool, new_reservation.start)
	# A change in end time will always be provided for reservation move and resize operations.
	new_reservation.end = reservation_to_cancel.end + end_delta
	new_reservation.tool = reservation_to_cancel.tool
	new_reservation.project = reservation_to_cancel.project
	new_reservation.user = reservation_to_cancel.user
	new_reservation.creation_time = now
	policy_problems, overridable = check_policy_to_save_reservation(reservation_to_cancel, new_reservation, request.user, False)
	if policy_problems:
		return HttpResponseBadRequest(policy_problems[0])
	else:
		# All policy checks passed, so save the reservation.
		new_reservation.save_and_notify()
		reservation_to_cancel.descendant = new_reservation
		reservation_to_cancel.save_and_notify()
	return response


def modify_outage(request, start_delta, end_delta):
	try:
		outage = ScheduledOutage.objects.get(pk=request.POST.get('id'))
	except ScheduledOutage.DoesNotExist:
		return HttpResponseNotFound("The outage that you wish to modify doesn't exist!")
	if start_delta:
		outage.start += start_delta
	outage.end += end_delta
	policy_problem = check_policy_to_create_outage(outage)
	if policy_problem:
		return HttpResponseBadRequest(policy_problem)
	else:
		# All policy checks passed, so save the reservation.
		outage.save()
	return HttpResponse()


def determine_insufficient_notice(tool, start):
	""" Determines if a reservation is created that does not give the
	NanoFab staff sufficient advance notice to configure a tool. """
	for config in tool.configuration_set.all():
		advance_notice = start - timezone.now()
		if advance_notice < timedelta(hours=config.advance_notice_limit):
			return True
	return False


@login_required
@require_POST
def cancel_reservation(request, reservation_id):
	""" Cancel a reservation for a user. """
	reservation = get_object_or_404(Reservation, id=reservation_id)

	reason = parse_parameter_string(request.POST, 'reason')
	response = cancel_the_reservation(reservation=reservation, user=request.user, reason=reason)

	if request.device == 'desktop':
		return response
	if request.device == 'mobile':
		if response.status_code == HTTPStatus.OK:
			return render(request, 'mobile/cancellation_result.html', {'event_type': 'Reservation', 'tool': reservation.tool})
		else:
			return render(request, 'mobile/error.html', {'message': response.content})


@staff_member_required(login_url=None)
@require_POST
def cancel_outage(request, outage_id):
	outage = get_object_or_404(ScheduledOutage, id=outage_id)
	outage.delete()
	if request.device == 'desktop':
		return HttpResponse()
	if request.device == 'mobile':
		dictionary = {'event_type': 'Scheduled outage', 'tool': outage.tool}
		return render(request, 'mobile/cancellation_result.html', dictionary)


@staff_member_required(login_url=None)
@require_POST
def set_reservation_title(request, reservation_id):
	""" Change reservation title for a user. """
	reservation = get_object_or_404(Reservation, id=reservation_id)
	reservation.title = request.POST.get('title', '')[:reservation._meta.get_field('title').max_length]
	reservation.save()
	return HttpResponse()


@login_required
@permission_required('NEMO.trigger_timed_services', raise_exception=True)
@require_GET
def email_reservation_reminders(request):
	# Exit early if the reservation reminder email template has not been customized for the organization yet.
	reservation_reminder_message = get_media_file_contents('reservation_reminder_email.html')
	reservation_warning_message = get_media_file_contents('reservation_warning_email.html')
	if not reservation_reminder_message or not reservation_warning_message:
		return HttpResponseNotFound('The reservation reminder email template has not been customized for your organization yet. Please visit the NEMO customizable_key_values page to upload a template, then reservation reminder email notifications can be sent.')

	# Find all reservations that are two hours from now, plus or minus 5 minutes to allow for time skew.
	preparation_time = 120
	tolerance = 5
	earliest_start = timezone.now() + timedelta(minutes=preparation_time) - timedelta(minutes=tolerance)
	latest_start = timezone.now() + timedelta(minutes=preparation_time) + timedelta(minutes=tolerance)
	upcoming_reservations = Reservation.objects.filter(cancelled=False, start__gt=earliest_start, start__lt=latest_start)
	# Email a reminder to each user with an upcoming reservation.
	for reservation in upcoming_reservations:
		tool = reservation.tool
		if tool.operational and not tool.problematic() and tool.all_resources_available():
			subject = reservation.tool.name + " reservation reminder"
			rendered_message = Template(reservation_reminder_message).render(Context({'reservation': reservation, 'template_color': bootstrap_primary_color('success')}))
		elif not tool.operational or tool.required_resource_is_unavailable():
			subject = reservation.tool.name + " reservation problem"
			rendered_message = Template(reservation_warning_message).render(Context({'reservation': reservation, 'template_color': bootstrap_primary_color('danger'), 'fatal_error': True}))
		else:
			subject = reservation.tool.name + " reservation warning"
			rendered_message = Template(reservation_warning_message).render(Context({'reservation': reservation, 'template_color': bootstrap_primary_color('warning'), 'fatal_error': False}))
		user_office_email = get_customization('user_office_email_address')
		reservation.user.email_user(subject, rendered_message, user_office_email)
	return HttpResponse()


@login_required
@permission_required('NEMO.trigger_timed_services', raise_exception=True)
@require_GET
def email_usage_reminders(request):
	projects_to_exclude = request.GET.getlist("projects_to_exclude[]")
	busy_users = AreaAccessRecord.objects.filter(end=None, staff_charge=None).exclude(project__id__in=projects_to_exclude)
	busy_tools = UsageEvent.objects.filter(end=None).exclude(project__id__in=projects_to_exclude)

	# Make lists of all the things a user is logged in to.
	# We don't want to send 3 separate emails if a user is logged into three things.
	# Just send one email for all the things!
	aggregate = {}
	for access_record in busy_users:
		key = str(access_record.customer)
		aggregate[key] = {
			'email': access_record.customer.email,
			'first_name': access_record.customer.first_name,
			'resources_in_use': [str(access_record.area)],
		}
	for usage_event in busy_tools:
		key = str(usage_event.operator)
		if key in aggregate:
			aggregate[key]['resources_in_use'].append(usage_event.tool.name)
		else:
			aggregate[key] = {
				'email': usage_event.operator.email,
				'first_name': usage_event.operator.first_name,
				'resources_in_use': [usage_event.tool.name],
			}

	user_office_email = get_customization('user_office_email_address')

	message = get_media_file_contents('usage_reminder_email.html')
	if message:
		subject = "NanoFab usage"
		for user in aggregate.values():
			rendered_message = Template(message).render(Context({'user': user}))
			send_mail(subject, rendered_message, user_office_email, [user['email']])

	message = get_media_file_contents('staff_charge_reminder_email.html')
	if message:
		busy_staff = StaffCharge.objects.filter(end=None)
		for staff_charge in busy_staff:
			subject = "Active staff charge since " + format_datetime(staff_charge.start)
			rendered_message = Template(message).render(Context({'staff_charge': staff_charge}))
			staff_charge.staff_member.email_user(subject, rendered_message, user_office_email)

	return HttpResponse()


@login_required
@require_GET
def reservation_details(request, reservation_id):
	reservation = get_object_or_404(Reservation, id=reservation_id)
	if reservation.cancelled:
		error_message = 'This reservation was cancelled by {0} at {1}.'.format(reservation.cancelled_by, format_datetime(reservation.cancellation_time))
		return HttpResponseNotFound(error_message)
	return render(request, 'calendar/reservation_details.html', {'reservation': reservation})


@login_required
@require_GET
def outage_details(request, outage_id):
	outage = get_object_or_404(ScheduledOutage, id=outage_id)
	return render(request, 'calendar/outage_details.html', {'outage': outage})


@login_required
@require_GET
def usage_details(request, event_id):
	event = get_object_or_404(UsageEvent, id=event_id)
	return render(request, 'calendar/usage_details.html', {'event': event})


@login_required
@require_GET
def area_access_details(request, event_id):
	event = get_object_or_404(AreaAccessRecord, id=event_id)
	return render(request, 'calendar/area_access_details.html', {'event': event})


@login_required
@require_GET
@permission_required('NEMO.trigger_timed_services', raise_exception=True)
def cancel_unused_reservations(request):
	# Exit early if the missed reservation email template has not been customized for the organization yet.
	if not get_media_file_contents('missed_reservation_email.html'):
		return HttpResponseNotFound('The missed reservation email template has not been customized for your organization yet. Please visit the NEMO customizable_key_values page to upload a template, then missed email notifications can be sent.')

	tools = Tool.objects.filter(visible=True, _operational=True, _missed_reservation_threshold__isnull=False)
	missed_reservations = []
	for tool in tools:
		# If a tool is in use then there's no need to look for unused reservation time.
		if tool.in_use() or tool.required_resource_is_unavailable() or tool.scheduled_outage_in_progress():
			continue
		# Calculate the timestamp of how long a user can be late for a reservation.
		threshold = (timezone.now() - timedelta(minutes=tool.missed_reservation_threshold))
		threshold = datetime.replace(threshold, second=0, microsecond=0)  # Round down to the nearest minute.
		# Find the reservations that began exactly at the threshold.
		reservation = Reservation.objects.filter(cancelled=False, missed=False, shortened=False, tool=tool, user__is_staff=False, start=threshold, end__gt=timezone.now())
		for r in reservation:
			# Staff may abandon reservations.
			if r.user.is_staff:
				continue
			# If there was no tool enable or disable event since the threshold timestamp then we assume the reservation has been missed.
			if not (UsageEvent.objects.filter(tool_id__in=tool.get_family_tool_ids(), start__gte=threshold).exists() or UsageEvent.objects.filter(tool_id__in=tool.get_family_tool_ids(), end__gte=threshold).exists()):
				# Mark the reservation as missed and notify the user & NanoFab staff.
				r.missed = True
				r.save()
				missed_reservations.append(r)

	for r in missed_reservations:
		send_missed_reservation_notification(r)

	return HttpResponse()


@staff_member_required(login_url=None)
@require_GET
def proxy_reservation(request):
	return render(request, 'calendar/proxy_reservation.html', {'users': User.objects.filter(is_active=True)})


def cancel_the_reservation(reservation: Reservation, user: User, reason: Union[str, None]):
	response = check_policy_to_cancel_reservation(reservation, user)
	# Staff must provide a reason when cancelling a reservation they do not own.
	if reservation.user != user and not reason:
		response = HttpResponseBadRequest("You must provide a reason when cancelling someone else's reservation.")

	if response.status_code == HTTPStatus.OK:
		# All policy checks passed, so cancel the reservation.
		reservation.cancelled = True
		reservation.cancellation_time = timezone.now()
		reservation.cancelled_by = user

		if reason:
			''' don't notify in this case since we are sending a specific email for the cancellation '''
			reservation.save()
			dictionary = {
				'staff_member': user,
				'reservation': reservation,
				'reason': reason,
				'template_color': bootstrap_primary_color('info')
			}
			email_contents = get_media_file_contents('cancellation_email.html')
			if email_contents:
				cancellation_email = Template(email_contents).render(Context(dictionary))
				if getattr(reservation.user.preferences, 'attach_cancelled_reservation', False):
					attachment = create_ics_for_reservation(reservation, cancelled=True)
					reservation.user.email_user('Your reservation was cancelled', cancellation_email, user.email, [attachment])
				else:
					reservation.user.email_user('Your reservation was cancelled', cancellation_email, user.email)

		else:
			''' here the user cancelled his own reservation so notify him '''
			reservation.save_and_notify()

	return response


def send_missed_reservation_notification(reservation):
	subject = "Missed reservation for the " + str(reservation.tool)
	message = get_media_file_contents('missed_reservation_email.html')
	message = Template(message).render(Context({'reservation': reservation}))
	user_office_email = get_customization('user_office_email_address')
	abuse_email = get_customization('abuse_email_address')
	send_mail(subject, message, user_office_email, [reservation.user.email, abuse_email, user_office_email])


def send_user_created_reservation_notification(reservation: Reservation):
	if getattr(reservation.user.preferences, 'attach_created_reservation', False):
		subject = "[NEMO] Reservation for the " + str(reservation.tool)
		message = get_media_file_contents('reservation_created_user_email.html')
		message = Template(message).render(Context({'reservation': reservation}))
		user_office_email = get_customization('user_office_email_address')
		attachment = create_ics_for_reservation(reservation)
		reservation.user.email_user(subject, message, user_office_email, [attachment])


def send_user_cancelled_reservation_notification(reservation: Reservation):
	if getattr(reservation.user.preferences, 'attach_cancelled_reservation', False):
		subject = "[NEMO] Cancelled Reservation for the " + str(reservation.tool)
		message = get_media_file_contents('reservation_cancelled_user_email.html')
		message = Template(message).render(Context({'reservation': reservation}))
		user_office_email = get_customization('user_office_email_address')
		attachment = create_ics_for_reservation(reservation, cancelled=True)
		reservation.user.email_user(subject, message, user_office_email, [attachment])


def create_ics_for_reservation(reservation: Reservation, cancelled=False):
	method = 'METHOD:CANCEL\n' if cancelled else 'METHOD:PUBLISH\n'
	status = 'STATUS:CANCELLED\n' if cancelled else 'STATUS:CONFIRMED\n'
	uid = 'UID:'+str(reservation.id)+'\n'
	sequence = 'SEQUENCE:2\n' if cancelled else 'SEQUENCE:0\n'
	priority = 'PRIORITY:5\n' if cancelled else 'PRIORITY:0\n'
	now = datetime.now().strftime('%Y%m%dT%H%M%S')
	start = reservation.start.astimezone(timezone.get_current_timezone()).strftime('%Y%m%dT%H%M%S')
	end = reservation.end.astimezone(timezone.get_current_timezone()).strftime('%Y%m%dT%H%M%S')
	lines = ['BEGIN:VCALENDAR\n', 'VERSION:2.0\n', method, 'BEGIN:VEVENT\n', uid, sequence, priority, f'DTSTAMP:{now}\n', f'DTSTART:{start}\n', f'DTEND:{end}\n', f'SUMMARY:[NEMO] {reservation.tool.name} Reservation\n', status, 'END:VEVENT\n', 'END:VCALENDAR\n']
	ics = io.StringIO('')
	ics.writelines(lines)
	ics.seek(0)

	filename = 'cancelled_nemo_reservation.ics' if cancelled else 'nemo_reservation.ics'

	return create_email_attachment(ics, filename)



from datetime import datetime
from typing import List, Dict

from django import forms
from django.utils import timezone
from rest_framework import status
from rest_framework.decorators import api_view
from rest_framework.response import Response
from rest_framework.viewsets import ReadOnlyModelViewSet

from NEMO.filters import ReservationFilter, UsageEventFilter, AreaAccessRecordFilter, UserFilter
from NEMO.models import User, Project, Account, Reservation, UsageEvent, AreaAccessRecord, Task, ScheduledOutage, Tool, \
	ConsumableWithdraw
from NEMO.serializers import UserSerializer, ProjectSerializer, AccountSerializer, ReservationSerializer, \
	UsageEventSerializer, AreaAccessRecordSerializer, TaskSerializer, ScheduledOutageSerializer, ToolSerializer, \
	BillableItemSerializer
from NEMO.utilities import localize


date_time_format = '%m/%d/%Y %H:%M:%S'

class BillingFilterForm(forms.Form):
	start = forms.DateField(required=True)
	end = forms.DateField(required=True)
	username = forms.CharField(required=False)
	account_name = forms.CharField(required=False)
	account_id = forms.IntegerField(required=False)
	application_name = forms.CharField(required=False)
	project_name = forms.CharField(required=False)
	project_id = forms.IntegerField(required=False)

	def get_start_date(self):
		return localize(datetime.combine(self.cleaned_data['start'], datetime.min.time()))

	def get_end_date(self):
		return localize(datetime.combine(self.cleaned_data['end'], datetime.max.time()))

	def get_project_id(self):
		return self.cleaned_data['project_id']

	def get_project_name(self):
		return self.cleaned_data['project_name']

	def get_account_id(self):
		return self.cleaned_data['account_id']

	def get_account_name(self):
		return self.cleaned_data['account_name']

	def get_username(self):
		return self.cleaned_data['username']

	def get_application_name(self):
		return self.cleaned_data['application_name']


class UserViewSet(ReadOnlyModelViewSet):
	queryset = User.objects.all()
	serializer_class = UserSerializer
	filter_class = UserFilter


class ProjectViewSet(ReadOnlyModelViewSet):
	queryset = Project.objects.all()
	serializer_class = ProjectSerializer


class AccountViewSet(ReadOnlyModelViewSet):
	queryset = Account.objects.all()
	serializer_class = AccountSerializer


class ToolViewSet(ReadOnlyModelViewSet):
	queryset = Tool.objects.all()
	serializer_class = ToolSerializer


class ReservationViewSet(ReadOnlyModelViewSet):
	queryset = Reservation.objects.all()
	serializer_class = ReservationSerializer
	filter_class = ReservationFilter


class UsageEventViewSet(ReadOnlyModelViewSet):
	queryset = UsageEvent.objects.all()
	serializer_class = UsageEventSerializer
	filter_class = UsageEventFilter


class AreaAccessRecordViewSet(ReadOnlyModelViewSet):
	queryset = AreaAccessRecord.objects.all()
	serializer_class = AreaAccessRecordSerializer
	filter_class = AreaAccessRecordFilter


class TaskViewSet(ReadOnlyModelViewSet):
	queryset = Task.objects.all()
	serializer_class = TaskSerializer


class ScheduledOutageViewSet(ReadOnlyModelViewSet):
	queryset = ScheduledOutage.objects.all()
	serializer_class = ScheduledOutageSerializer


billing = [UsageEvent, AreaAccessRecord, 'Consumable', 'Missed Reservation']


@api_view(["GET"])
def billing(request):
	form = BillingFilterForm(request.GET)
	if not form.is_valid():
		return Response(status=status.HTTP_400_BAD_REQUEST, data=form.errors)

	data :List[Dict] = []
	usage_events = get_usage_events_for_billing(form)
	area_access = get_area_access_for_billing(form)
	consumables = get_consumables_for_billing(form)
	missed_reservation = get_missed_reservations_for_billing(form)

	data.extend(usage_events)
	data.extend(area_access)
	data.extend(consumables)
	data.extend(missed_reservation)

	serializer = BillableItemSerializer(data, many=True)
	return Response(serializer.data)


def get_usage_events_for_billing(billing_form: BillingFilterForm) -> List[Dict]:
	result = []
	queryset = UsageEvent.objects.filter()
	start, end = billing_form.get_start_date(), billing_form.get_end_date()
	queryset = queryset.filter(start__gte=start, end__lte=end, start__lte=end, end__gte=start)
	if billing_form.get_account_id():
		queryset = queryset.filter(project__account_id=billing_form.get_account_id())
	if billing_form.get_account_name():
		queryset = queryset.filter(project__account__name=billing_form.get_account_name())
	if billing_form.get_project_id():
		queryset = queryset.filter(project__id=billing_form.get_project_id())
	if billing_form.get_project_name():
		queryset = queryset.filter(project__name=billing_form.get_project_name())
	if billing_form.get_application_name():
		queryset = queryset.filter(project__application_identifier=billing_form.get_application_name())
	if billing_form.get_username():
		queryset = queryset.filter(user__username=billing_form.get_username())
	for usage_event in queryset:
		diff = usage_event.end - usage_event.start
		result.append({
			'type': 'tool_usage',
			'details': usage_event.tool.name,
			'account': usage_event.project.account.name,
			'account_id': usage_event.project.account_id,
			'project': usage_event.project.name,
			'project_id': usage_event.project_id,
			'application': usage_event.project.application_identifier,
			'username': usage_event.user.username,
			'user_id': usage_event.user_id,
			'start': usage_event.start.astimezone(timezone.get_current_timezone()).strftime(date_time_format),
			'end': usage_event.end.astimezone(timezone.get_current_timezone()).strftime(date_time_format),
			'quantity': str(round(diff.days*1440 + diff.seconds/60, 2))
		})
	return result

def get_area_access_for_billing(billing_form: BillingFilterForm) -> List[Dict]:
	result = []
	queryset = AreaAccessRecord.objects.filter()
	start, end = billing_form.get_start_date(), billing_form.get_end_date()
	queryset = queryset.filter(start__gte=start, end__lte=end, start__lte=end, end__gte=start)
	if billing_form.get_account_id():
		queryset = queryset.filter(project__account_id=billing_form.get_account_id())
	if billing_form.get_account_name():
		queryset = queryset.filter(project__account__name=billing_form.get_account_name())
	if billing_form.get_project_id():
		queryset = queryset.filter(project__id=billing_form.get_project_id())
	if billing_form.get_project_name():
		queryset = queryset.filter(project__name=billing_form.get_project_name())
	if billing_form.get_application_name():
		queryset = queryset.filter(project__application_identifier=billing_form.get_application_name())
	if billing_form.get_username():
		queryset = queryset.filter(customer__username=billing_form.get_username())
	for area_access_record in queryset:
		diff = area_access_record.end - area_access_record.start
		result.append({
			'type': 'area_acess',
			'details': area_access_record.area.name,
			'account': area_access_record.project.account.name,
			'account_id': area_access_record.project.account_id,
			'project': area_access_record.project.name,
			'project_id': area_access_record.project_id,
			'application': area_access_record.project.application_identifier,
			'username': area_access_record.customer.username,
			'user_id': area_access_record.customer_id,
			'start': area_access_record.start.astimezone(timezone.get_current_timezone()).strftime(date_time_format),
			'end': area_access_record.end.astimezone(timezone.get_current_timezone()).strftime(date_time_format),
			'quantity': str(round(diff.days*1440 + diff.seconds/60, 2))
		})
	return result


def get_missed_reservations_for_billing(billing_form: BillingFilterForm) -> List[Dict]:
	result = []
	queryset = Reservation.objects.filter(missed=True)
	start, end = billing_form.get_start_date(), billing_form.get_end_date()
	queryset = queryset.filter(start__gte=start, end__lte=end, start__lte=end, end__gte=start)
	if billing_form.get_account_id():
		queryset = queryset.filter(project__account_id=billing_form.get_account_id())
	if billing_form.get_account_name():
		queryset = queryset.filter(project__account__name=billing_form.get_account_name())
	if billing_form.get_project_id():
		queryset = queryset.filter(project__id=billing_form.get_project_id())
	if billing_form.get_project_name():
		queryset = queryset.filter(project__name=billing_form.get_project_name())
	if billing_form.get_application_name():
		queryset = queryset.filter(project__application_identifier=billing_form.get_application_name())
	if billing_form.get_username():
		queryset = queryset.filter(user__username=billing_form.get_username())
	for missed_reservation in queryset:
		result.append({
			'type': 'missed_reservation',
			'details': missed_reservation.tool.name,
			'account': missed_reservation.project.account.name,
			'account_id': missed_reservation.project.account_id,
			'project': missed_reservation.project.name,
			'project_id': missed_reservation.project_id,
			'application': missed_reservation.project.application_identifier,
			'username': missed_reservation.user.username,
			'user_id': missed_reservation.user_id,
			'start': missed_reservation.start.astimezone(timezone.get_current_timezone()).strftime(date_time_format),
			'end': missed_reservation.end.astimezone(timezone.get_current_timezone()).strftime(date_time_format),
			'quantity': 1
		})
	return result


def get_consumables_for_billing(billing_form: BillingFilterForm) -> List[Dict]:
	result = []
	queryset = ConsumableWithdraw.objects.filter()
	start, end = billing_form.get_start_date(), billing_form.get_end_date()
	queryset = queryset.filter(date__gte=start, date__lte=end)
	if billing_form.get_account_id():
		queryset = queryset.filter(project__account_id=billing_form.get_account_id())
	if billing_form.get_account_name():
		queryset = queryset.filter(project__account__name=billing_form.get_account_name())
	if billing_form.get_project_id():
		queryset = queryset.filter(project__id=billing_form.get_project_id())
	if billing_form.get_project_name():
		queryset = queryset.filter(project__name=billing_form.get_project_name())
	if billing_form.get_application_name():
		queryset = queryset.filter(project__application_identifier=billing_form.get_application_name())
	if billing_form.get_username():
		queryset = queryset.filter(customer__username=billing_form.get_username())
	for consumable_withdrawal in queryset:
		result.append({
			'type': 'consumable',
			'details': consumable_withdrawal.consumable.name,
			'account': consumable_withdrawal.project.account.name,
			'account_id': consumable_withdrawal.project.account_id,
			'project': consumable_withdrawal.project.name,
			'project_id': consumable_withdrawal.project_id,
			'application': consumable_withdrawal.project.application_identifier,
			'username': consumable_withdrawal.customer.username,
			'user_id': consumable_withdrawal.customer_id,
			'start': consumable_withdrawal.date.astimezone(timezone.get_current_timezone()).strftime(date_time_format),
			'end': consumable_withdrawal.date.astimezone(timezone.get_current_timezone()).strftime(date_time_format),
			'quantity': consumable_withdrawal.quantity
		})
	return result



from datetime import timedelta

from django.contrib.contenttypes.models import ContentType
from django.utils import timezone

from NEMO.models import News, Notification, SafetyIssue, User


def delete_expired_notifications():
	Notification.objects.filter(expiration__lt=timezone.now()).delete()


def get_notifications(user, notification_type):
	content_type = ContentType.objects.get_for_model(notification_type)
	notifications = Notification.objects.filter(user=user, content_type=content_type)
	if notifications:
		notification_ids = list(notifications.values_list('object_id', flat=True))
		notifications.delete()
		return notification_ids
	else:
		return None


def get_notificaiton_counts(user):
	counts = {}
	for t in Notification.Types.Choices:
		model = t[0]
		content_type = ContentType.objects.get(app_label='NEMO', model=model)
		counts[model] = Notification.objects.filter(user=user, content_type=content_type).count()
	return counts


def create_news_notification(story):
	content_type = ContentType.objects.get_for_model(News)
	Notification.objects.filter(content_type=content_type, object_id=story.id).delete()  # Delete all existing notifications for this story, so we don't have multiple notifications for the same story
	users = User.objects.filter(is_active=True)
	expiration = timezone.now() + timedelta(days=30)  # Unread news story notifications always expire after 30 days
	for u in users:
		Notification.objects.create(user=u, expiration=expiration, content_object=story)


def delete_news_notification(story):
	content_type = ContentType.objects.get_for_model(News)
	Notification.objects.filter(content_type=content_type, object_id=story.id).delete()


def create_safety_notification(safety_issue):
	users = User.objects.filter(is_staff=True, is_active=True)
	expiration = timezone.now() + timedelta(days=30)  # Unread safety issue notifications always expire after 30 days
	for u in users:
		Notification.objects.create(user=u, expiration=expiration, content_object=safety_issue)


def delete_safety_notification(issue):
	content_type = ContentType.objects.get_for_model(SafetyIssue)
	Notification.objects.filter(content_type=content_type, object_id=issue.id).delete()



from logging import getLogger
from threading import Thread

postponed_tasks_logger = getLogger("NEMO.PostponedTasks")


def postpone(function):
	def decorator(*arguments, **named_arguments):
		t = Thread(target=function, args=arguments, kwargs=named_arguments)
		t.daemon = True
		t.start()
	return decorator



from django import forms
from django.contrib import admin
from django.contrib.admin import register
from django.contrib.admin.widgets import FilteredSelectMultiple
from django.contrib.auth.models import Permission

from NEMO.actions import lock_selected_interlocks, synchronize_with_tool_usage, unlock_selected_interlocks
from NEMO.models import Account, ActivityHistory, Alert, Area, AreaAccessRecord, Comment, Configuration, \
	ConfigurationHistory, Consumable, ConsumableCategory, ConsumableWithdraw, ContactInformation, \
	ContactInformationCategory, Customization, Door, Interlock, InterlockCard, LandingPageChoice, MembershipHistory, \
	News, Notification, PhysicalAccessLevel, PhysicalAccessLog, Project, Reservation, Resource, ResourceCategory, \
	SafetyIssue, ScheduledOutage, ScheduledOutageCategory, StaffCharge, Task, TaskCategory, TaskHistory, TaskStatus, \
	Tool, TrainingSession, UsageEvent, User, UserType, UserPreferences, TaskImages, InterlockCardCategory

admin.site.site_header = "NEMO"
admin.site.site_title = "NEMO"
admin.site.index_title = "Detailed administration"


def record_remote_many_to_many_changes_and_save(request, obj, form, change, many_to_many_field, save_function_pointer):
	"""
	Record the changes in a many-to-many field that the model does not own. Then, save the many-to-many field.
	"""
	# If the model object is being changed then we can get the list of previous members.
	if change:
		original_members = set(obj.user_set.all())
	else:  # The model object is being created (instead of changed) so we can assume there are no members (initially).
		original_members = set()
	current_members = set(form.cleaned_data[many_to_many_field])
	added_members = []
	removed_members = []

	# Log membership changes if they occurred.
	symmetric_difference = original_members ^ current_members
	if symmetric_difference:
		if change:  # the members have changed, so find out what was added and removed...
			# We can can see the previous members of the object model by looking it up
			# in the database because the member list hasn't been committed yet.
			added_members = set(current_members) - set(original_members)
			removed_members = set(original_members) - set(current_members)

		else:  # a model object is being created (instead of changed) so we can assume all the members are new...
			added_members = form.cleaned_data[many_to_many_field]

	# A primary key for the object is required to make many-to-many field changes.
	# If the object is being changed then it has already been assigned a primary key.
	if not change:
		save_function_pointer(request, obj, form, change)
	obj.user_set = form.cleaned_data[many_to_many_field]
	save_function_pointer(request, obj, form, change)

	# Record which members were added to the object.
	for user in added_members:
		new_member = MembershipHistory()
		new_member.authorizer = request.user
		new_member.parent_content_object = obj
		new_member.child_content_object = user
		new_member.action = MembershipHistory.Action.ADDED
		new_member.save()

	# Record which members were removed from the object.
	for user in removed_members:
		ex_member = MembershipHistory()
		ex_member.authorizer = request.user
		ex_member.parent_content_object = obj
		ex_member.child_content_object = user
		ex_member.action = MembershipHistory.Action.REMOVED
		ex_member.save()


def record_local_many_to_many_changes(request, obj, form, many_to_many_field):
	"""
	Record the changes in a many-to-many field that the model owns.
	"""
	if many_to_many_field in form.changed_data:
		original_members = set(getattr(obj, many_to_many_field).all())
		current_members = set(form.cleaned_data[many_to_many_field])
		added_members = set(current_members) - set(original_members)
		for a in added_members:
			p = MembershipHistory()
			p.action = MembershipHistory.Action.ADDED
			p.authorizer = request.user
			p.child_content_object = obj
			p.parent_content_object = a
			p.save()
		removed_members = set(original_members) - set(current_members)
		for a in removed_members:
			p = MembershipHistory()
			p.action = MembershipHistory.Action.REMOVED
			p.authorizer = request.user
			p.child_content_object = obj
			p.parent_content_object = a
			p.save()


def record_active_state(request, obj, form, field_name, is_initial_creation):
	"""
	Record whether the account, project, or user is active when the active state is changed.
	"""
	if field_name in form.changed_data or is_initial_creation:
		activity_entry = ActivityHistory()
		activity_entry.authorizer = request.user
		activity_entry.action = getattr(obj, field_name)
		activity_entry.content_object = obj
		activity_entry.save()


class ToolAdminForm(forms.ModelForm):
	class Meta:
		model = Tool
		fields = '__all__'

	class Media:
		js = ("tool_form_admin.js",)

	qualified_users = forms.ModelMultipleChoiceField(
		queryset=User.objects.all(),
		required=False,
		widget=FilteredSelectMultiple(
			verbose_name='Users',
			is_stacked=False
		)
	)

	_backup_owners = forms.ModelMultipleChoiceField(
		queryset=User.objects.all(),
		required=False,
		widget=FilteredSelectMultiple(
			verbose_name='Users',
			is_stacked=False
		)
	)

	required_resources = forms.ModelMultipleChoiceField(
		queryset=Resource.objects.all(),
		required=False,
		widget=FilteredSelectMultiple(
			verbose_name='Required resources',
			is_stacked=False
		)
	)

	nonrequired_resources = forms.ModelMultipleChoiceField(
		queryset=Resource.objects.all(),
		required=False,
		widget=FilteredSelectMultiple(
			verbose_name='Nonrequired resources',
			is_stacked=False
		)
	)

	def __init__(self, *args, **kwargs):
		super(ToolAdminForm, self).__init__(*args, **kwargs)
		if self.instance.pk:
			self.fields['qualified_users'].initial = self.instance.user_set.all()
			self.fields['required_resources'].initial = self.instance.required_resource_set.all()
			self.fields['nonrequired_resources'].initial = self.instance.nonrequired_resource_set.all()


	def clean(self):
		cleaned_data = super().clean()
		parent_tool = cleaned_data.get("parent_tool")
		category = cleaned_data.get("_category")
		location = cleaned_data.get("_location")
		phone_number = cleaned_data.get("_phone_number")
		primary_owner = cleaned_data.get("_primary_owner")

		if parent_tool:
			if parent_tool.id == self.instance.id:
				self.add_error('parent_tool', 'You cannot select the parent to be the tool itself.')
			# in case of alternate tool, remove everything except parent_tool and name
			data = dict([(k, v) for k, v in self.cleaned_data.items() if k == "parent_tool" or k == "name"])
			# an alternate tool is never visible
			data['visible'] = False
			return data
		else:
			if not category:
				self.add_error('_category', 'This field is required.')
			if not location:
				self.add_error('_location', 'This field is required.')
			if not phone_number:
				self.add_error('_phone_number', 'This field is required.')
			if not primary_owner:
				self.add_error('_primary_owner', 'This field is required.')
			policy_off_between_times = cleaned_data.get("_policy_off_between_times")
			policy_off_start_time = cleaned_data.get("_policy_off_start_time")
			policy_off_end_time = cleaned_data.get("_policy_off_end_time")
			if policy_off_between_times and (not policy_off_start_time or not policy_off_end_time):
				if not policy_off_start_time:
					self.add_error("_policy_off_start_time", "Start time must be specified")
				if not policy_off_end_time:
					self.add_error("_policy_off_end_time", "End time must be specified")


@register(Tool)
class ToolAdmin(admin.ModelAdmin):
	list_display = ('name_display', '_category', 'visible', 'operational_display', 'problematic', 'is_configurable')
	list_filter = ('visible', '_operational', '_category')
	form = ToolAdminForm
	fieldsets = (
		(None, {'fields': ('name', 'parent_tool', '_category', 'qualified_users', '_post_usage_questions'),}),
		('Current state', {'fields': ('visible', '_operational'),}),
		('Contact information', {'fields': ('_primary_owner', '_backup_owners', '_notification_email_address', '_location', '_phone_number'),}),
		('Reservation', {'fields': ('_reservation_horizon', '_missed_reservation_threshold'),}),
		('Usage policy', {'fields': ('_policy_off_between_times', '_policy_off_start_time', '_policy_off_end_time', '_policy_off_weekend', '_minimum_usage_block_time', '_maximum_usage_block_time', '_maximum_reservations_per_day', '_minimum_time_between_reservations', '_maximum_future_reservation_time',),}),
		('Area Access', {'fields': ('_requires_area_access', '_grant_physical_access_level_upon_qualification', '_grant_badge_reader_access_upon_qualification', '_interlock', '_allow_delayed_logoff'),}),
		('Dependencies', {'fields': ('required_resources', 'nonrequired_resources'),}),
	)

	def formfield_for_foreignkey(self, db_field, request, **kwargs):
		""" We only want non children tool to be eligible as parents """
		if db_field.name == "parent_tool":
			kwargs["queryset"] = Tool.objects.filter(parent_tool__isnull=True)
		return super().formfield_for_foreignkey(db_field, request, **kwargs)

	def save_model(self, request, obj, form, change):
		"""
		Explicitly record any project membership changes.
		"""
		if obj.parent_tool:
			if obj.pk:
				# if this is an update (from regular to child tool), we want to make sure we are creating a clean version. In case the previous tool had fields that are now irrelevant
				clean_alt_tool = Tool(**form.cleaned_data)
				clean_alt_tool.pk = obj.pk
				obj = clean_alt_tool
			super(ToolAdmin, self).save_model(request, obj, form, change)
		else:
			record_remote_many_to_many_changes_and_save(request, obj, form, change, 'qualified_users', super(ToolAdmin, self).save_model)
			if 'required_resources' in form.changed_data:
				obj.required_resource_set = form.cleaned_data['required_resources']
			if 'nonrequired_resources' in form.changed_data:
				obj.nonrequired_resource_set = form.cleaned_data['nonrequired_resources']


@register(TrainingSession)
class TrainingSessionAdmin(admin.ModelAdmin):
	list_display = ('id', 'trainer', 'trainee', 'tool', 'project', 'type', 'date', 'duration', 'qualified')
	list_filter = ('qualified', 'date', 'type', 'tool')
	date_hierarchy = 'date'


@register(StaffCharge)
class StaffChargeAdmin(admin.ModelAdmin):
	list_display = ('id', 'staff_member', 'customer', 'start', 'end')
	list_filter = ('start',)
	date_hierarchy = 'start'


@register(AreaAccessRecord)
class AreaAccessRecordAdmin(admin.ModelAdmin):
	list_display = ('id', 'customer', 'area', 'project', 'start', 'end')
	list_filter = ('area', 'start',)
	date_hierarchy = 'start'


@register(Configuration)
class ConfigurationAdmin(admin.ModelAdmin):
	list_display = ('id', 'tool', 'name', 'qualified_users_are_maintainers', 'display_priority', 'exclude_from_configuration_agenda')
	filter_horizontal = ('maintainers',)


@register(ConfigurationHistory)
class ConfigurationHistoryAdmin(admin.ModelAdmin):
	list_display = ('id', 'configuration', 'user', 'modification_time', 'slot')
	date_hierarchy = 'modification_time'


@register(Account)
class AccountAdmin(admin.ModelAdmin):
	list_display = ('name', 'id', 'active')
	search_fields = ('name',)
	list_filter = ('active',)

	def save_model(self, request, obj, form, change):
		""" Audit account and project active status. """
		super(AccountAdmin, self).save_model(request, obj, form, change)
		record_active_state(request, obj, form, 'active', not change)


class ProjectAdminForm(forms.ModelForm):
	class Meta:
		model = Project
		fields = '__all__'

	members = forms.ModelMultipleChoiceField(
		queryset=User.objects.all(),
		required=False,
		widget=FilteredSelectMultiple(
			verbose_name='Users',
			is_stacked=False
		)
	)

	def __init__(self, *args, **kwargs):
		super(ProjectAdminForm, self).__init__(*args, **kwargs)
		if self.instance.pk:
			self.fields['members'].initial = self.instance.user_set.all()


@register(Project)
class ProjectAdmin(admin.ModelAdmin):
	list_display = ('name', 'id', 'application_identifier', 'account', 'active')
	search_fields = ('name', 'application_identifier', 'account__name')
	list_filter = ('active',)
	form = ProjectAdminForm

	def save_model(self, request, obj, form, change):
		"""
		Audit project creation and modification. Also save any project membership changes explicitly.
		"""
		record_remote_many_to_many_changes_and_save(request, obj, form, change, 'members', super(ProjectAdmin, self).save_model)
		# Make a history entry if a project has been moved under an account.
		# This applies to newly created projects and project ownership reassignment.
		if 'account' in form.changed_data:
			# Create a membership removal entry for the project if it used to belong to another account:
			if change:
				previous_account = MembershipHistory()
				previous_account.authorizer = request.user
				previous_account.child_content_object = obj
				previous_account.parent_content_object = Account.objects.get(pk=form.initial['account'])
				previous_account.action = MembershipHistory.Action.REMOVED
				previous_account.save()

			# Create a membership addition entry for the project with its current account.
			current_account = MembershipHistory()
			current_account.authorizer = request.user
			current_account.child_content_object = obj
			current_account.parent_content_object = obj.account
			current_account.action = MembershipHistory.Action.ADDED
			current_account.save()

		# Record whether the project is active or not.
		record_active_state(request, obj, form, 'active', not change)


@register(Reservation)
class ReservationAdmin(admin.ModelAdmin):
	list_display = ('id', 'user', 'creator', 'tool', 'project', 'start', 'end', 'duration', 'cancelled', 'missed')
	list_filter = ('cancelled', 'missed', 'tool')
	date_hierarchy = 'start'


@register(UsageEvent)
class UsageEventAdmin(admin.ModelAdmin):
	list_display = ('id', 'tool', 'user', 'operator', 'project', 'start', 'end', 'duration')
	list_filter = ('start', 'end', 'tool')
	date_hierarchy = 'start'


@register(Consumable)
class ConsumableAdmin(admin.ModelAdmin):
	list_display = ('name', 'quantity', 'category', 'visible', 'reminder_threshold', 'reminder_email')
	list_filter = ('visible', 'category')


@register(ConsumableCategory)
class ConsumableCategoryAdmin(admin.ModelAdmin):
	list_display = ('name',)


@register(ConsumableWithdraw)
class ConsumableWithdrawAdmin(admin.ModelAdmin):
	list_display = ('id', 'customer', 'merchant', 'consumable', 'quantity', 'project', 'date')
	list_filter = ('date', 'consumable')
	date_hierarchy = 'date'


class InterlockCardAdminForm(forms.ModelForm):
	class Meta:
		model = InterlockCard
		widgets = {
			'password': forms.PasswordInput(render_value=True),
		}
		fields = '__all__'

	def clean(self):
		if any(self.errors):
			return
		super(InterlockCardAdminForm, self).clean()
		category = self.cleaned_data['category']
		from NEMO import interlocks
		interlocks.get(category, False).clean_interlock_card(self)


@register(InterlockCard)
class InterlockCardAdmin(admin.ModelAdmin):
	form = InterlockCardAdminForm
	list_display = ('server', 'port', 'number', 'category', 'even_port', 'odd_port')


@register(Interlock)
class InterlockAdmin(admin.ModelAdmin):
	list_display = ('id', 'card', 'channel', 'state', 'tool', 'door')
	actions = [lock_selected_interlocks, unlock_selected_interlocks, synchronize_with_tool_usage]
	readonly_fields = ['state', 'most_recent_reply']


@register(InterlockCardCategory)
class InterlockCardCategoryAdmin(admin.ModelAdmin):
	list_display = ('name',)


@register(Task)
class TaskAdmin(admin.ModelAdmin):
	list_display = ('id', 'urgency', 'tool', 'creator', 'creation_time', 'problem_category', 'cancelled', 'resolved', 'resolution_category')
	list_filter = ('urgency', 'resolved', 'cancelled', 'safety_hazard', 'creation_time', 'tool')
	date_hierarchy = 'creation_time'


@register(TaskCategory)
class TaskCategoryAdmin(admin.ModelAdmin):
	list_display = ('name', 'stage')


@register(TaskStatus)
class TaskStatusAdmin(admin.ModelAdmin):
	list_display = ('name', 'notify_primary_tool_owner', 'notify_backup_tool_owners', 'notify_tool_notification_email', 'custom_notification_email_address')


@register(TaskHistory)
class TaskHistoryAdmin(admin.ModelAdmin):
	list_display = ('id', 'task', 'status', 'time', 'user')
	readonly_fields = ('time',)
	date_hierarchy = 'time'


@register(TaskImages)
class TaskImagesAdmin(admin.ModelAdmin):
	list_display = ('id', 'get_tool', 'task', 'uploaded_at')

	def get_tool(self, task_image: TaskImages):
		return task_image.task.tool.name
	get_tool.admin_order_field = 'tool'  # Allows column order sorting
	get_tool.short_description = 'Tool Name'  # Renames column head


@register(Comment)
class CommentAdmin(admin.ModelAdmin):
	list_display = ('id', 'tool', 'author', 'creation_date', 'expiration_date', 'visible', 'hidden_by', 'hide_date')
	list_filter = ('visible', 'creation_date', 'tool')
	date_hierarchy = 'creation_date'
	search_fields = ('content',)


@register(Resource)
class ResourceAdmin(admin.ModelAdmin):
	list_display = ('name', 'category', 'available')
	list_filter = ('available', 'category')
	filter_horizontal = ('fully_dependent_tools', 'partially_dependent_tools', 'dependent_areas')


@register(ActivityHistory)
class ActivityHistoryAdmin(admin.ModelAdmin):
	list_display = ('__str__', 'content_type', 'object_id', 'action', 'date', 'authorizer')
	date_hierarchy = 'date'


@register(MembershipHistory)
class MembershipHistoryAdmin(admin.ModelAdmin):
	list_display = ('__str__', 'parent_content_type', 'parent_object_id', 'action', 'child_content_type', 'child_object_id', 'date', 'authorizer')
	date_hierarchy = 'date'


@register(UserType)
class UserTypeAdmin(admin.ModelAdmin):
	list_display = ('name',)


@register(UserPreferences)
class UserPreferencesAdmin(admin.ModelAdmin):
	list_display = ('user',)


@register(User)
class UserAdmin(admin.ModelAdmin):
	filter_horizontal = ('groups', 'user_permissions', 'qualifications', 'projects', 'physical_access_levels')
	fieldsets = (
		('Personal information', {'fields': ('first_name', 'last_name', 'username', 'email', 'badge_number', 'type', 'domain')}),
		('Permissions', {'fields': ('is_active', 'is_staff', 'is_technician', 'is_superuser', 'training_required', 'groups', 'user_permissions', 'physical_access_levels')}),
		('Important dates', {'fields': ('date_joined', 'last_login', 'access_expiration')}),
		('NanoFab information', {'fields': ('qualifications', 'projects')}),
	)
	search_fields = ('first_name', 'last_name', 'username', 'email')
	list_display = ('first_name', 'last_name', 'username', 'email', 'is_active', 'domain', 'is_staff', 'is_technician', 'is_superuser', 'date_joined', 'last_login')
	list_filter = ('is_active', 'domain', 'is_staff', 'is_technician', 'is_superuser', 'date_joined', 'last_login')

	def formfield_for_manytomany(self, db_field, request, **kwargs):
		if db_field.name == "qualifications":
			kwargs["queryset"] = Tool.objects.filter(parent_tool__isnull=True)
		return super().formfield_for_manytomany(db_field, request, **kwargs)

	def save_model(self, request, obj, form, change):
		""" Audit project membership and qualifications when a user is saved. """
		super(UserAdmin, self).save_model(request, obj, form, change)
		record_local_many_to_many_changes(request, obj, form, 'projects')
		record_local_many_to_many_changes(request, obj, form, 'qualifications')
		record_local_many_to_many_changes(request, obj, form, 'physical_access_levels')
		record_active_state(request, obj, form, 'is_active', not change)


@register(PhysicalAccessLog)
class PhysicalAccessLogAdmin(admin.ModelAdmin):
	list_display = ('user', 'door', 'time', 'result')
	list_filter = ('door', 'result')
	search_fields = ('user',)
	date_hierarchy = 'time'


@register(SafetyIssue)
class SafetyIssueAdmin(admin.ModelAdmin):
	list_display = ('id', 'reporter', 'creation_time', 'visible', 'resolved', 'resolution_time', 'resolver')
	list_filter = ('resolved', 'visible', 'creation_time', 'resolution_time')
	readonly_fields = ('creation_time', 'resolution_time')
	search_fields = ('location', 'concern', 'progress', 'resolution',)


@register(Door)
class DoorAdmin(admin.ModelAdmin):
	list_display = ('name', 'area', 'interlock', 'get_absolute_url')


@register(Alert)
class AlertAdmin(admin.ModelAdmin):
	list_display = ('title', 'creation_time', 'creator', 'debut_time', 'expiration_time', 'user', 'dismissible')


@register(PhysicalAccessLevel)
class PhysicalAccessLevelAdmin(admin.ModelAdmin):
	list_display = ('name', 'area', 'schedule')


@register(ContactInformationCategory)
class ContactInformationCategoryAdmin(admin.ModelAdmin):
	list_display = ('name', 'display_order')


@register(ContactInformation)
class ContactInformationAdmin(admin.ModelAdmin):
	list_display = ('name', 'category')


@register(LandingPageChoice)
class LandingPageChoiceAdmin(admin.ModelAdmin):
	list_display = ('display_priority', 'name', 'url', 'open_in_new_tab', 'secure_referral', 'hide_from_mobile_devices', 'hide_from_desktop_computers')
	list_display_links = ('name',)


@register(Customization)
class CustomizationAdmin(admin.ModelAdmin):
	list_display = ('name', 'value')


@register(ScheduledOutageCategory)
class ScheduledOutageCategoryAdmin(admin.ModelAdmin):
	list_display = ('name',)


@register(ScheduledOutage)
class ScheduledOutageAdmin(admin.ModelAdmin):
	list_display = ('id', 'tool', 'resource', 'creator', 'title', 'start', 'end')


@register(News)
class NewsAdmin(admin.ModelAdmin):
	list_display = ('id', 'created', 'last_updated', 'archived', 'title')
	list_filter = ('archived',)


@register(Notification)
class NotificationAdmin(admin.ModelAdmin):
	list_display = ('id', 'user', 'expiration', 'content_type', 'object_id')


admin.site.register(ResourceCategory)
admin.site.register(Area)
admin.site.register(Permission)



from datetime import datetime

from django.contrib.auth.models import Permission
from django.test import TestCase
from django.utils import timezone

from NEMO.models import User, Tool, UsageEvent, Project, Account
from NEMO.tests.test_utilities import login_as_staff
from NEMO.views import api


class BillingAPITestCase(TestCase):

	def setUp(self):
		# create a few usage events etc.
		owner1 = User.objects.create(username='mctest1', first_name='Testy', last_name='McTester')
		tool1 = Tool.objects.create(name='test_tool1', primary_owner=owner1)
		owner2 = User.objects.create(username='mctest2', first_name='Testy', last_name='McTester')
		tool2 = Tool.objects.create(name='test_tool2', primary_owner=owner2)

		account = Account.objects.create(name="Test Account")
		project = Project.objects.create(name="Test Project", account=account, application_identifier='N19.0001')
		UsageEvent.objects.create(operator=owner1, user=owner1, tool=tool1, project=project, end=datetime.now().astimezone(timezone.get_current_timezone()))
		UsageEvent.objects.create(operator=owner2, user=owner2, tool=tool1, project=project, end=datetime.now().astimezone(timezone.get_current_timezone()))
		UsageEvent.objects.create(operator=owner2, user=owner2, tool=tool2, project=project, end=datetime.now().astimezone(timezone.get_current_timezone()))

	def test_billing_by_username(self):
		self.billing_by_attribute('username', 'mctest2', 'username', 2)


	def test_billing_by_account_name(self):
		self.billing_by_attribute('account_name', 'Test Account', 'account', 3)


	def test_billing_by_account_id(self):
		self.billing_by_attribute('account_id', 1, 'account_id', 3)


	def test_billing_by_project_name(self):
		self.billing_by_attribute('project_name', 'Test Project', 'project', 3)


	def test_billing_by_project_id(self):
		self.billing_by_attribute('project_id', 1, 'project_id', 3)


	def test_billing_by_application_name(self):
		self.billing_by_attribute('application_name', 'N19.0001', 'application', 3)


	def billing_by_attribute(self, attribute_name, attribute_value, result_attribute_name, results_number):
		data = {
			'start': datetime.now().strftime('%m/%d/%Y'),
			'end': datetime.now().strftime('%m/%d/%Y'),
			attribute_name: attribute_value
		}
		login_as_staff(self.client)

		response = self.client.get('/api/billing', data, follow=True)
		self.assertEquals(response.status_code, 403, "regular user or staff doesn't have permission")

		staff_user = User.objects.get(username='test_staff')
		staff_user.user_permissions.add(Permission.objects.get(codename='use_billing_api'))
		staff_user.save()
		response = self.client.get('/api/billing', data, follow=True)
		self.assertEquals(response.status_code, 200)
		self.assertEquals(len(response.data), results_number)

		date_format = api.date_time_format
		for billing_item in response.data:
			start = datetime.strptime(billing_item['start'], date_format)
			end = datetime.strptime(billing_item['end'], date_format)
			self.assertEqual(datetime.now().date(), start.date())
			self.assertEqual(datetime.now().date(), end.date())
			self.assertEqual(billing_item.get(result_attribute_name), attribute_value)




from collections import defaultdict

from django.contrib.admin.views.decorators import staff_member_required
from django.core.exceptions import ValidationError
from django.http import HttpResponseBadRequest
from django.shortcuts import render
from django.views.decorators.http import require_GET

from NEMO.forms import ReservationAbuseForm
from NEMO.models import Reservation, Tool, User


@staff_member_required(login_url=None)
@require_GET
def abuse(request):
	dictionary = {'tools': Tool.objects.filter(visible=True)}
	try:
		form = ReservationAbuseForm(request.GET)
		if form.is_valid():
			intermediate_results = defaultdict(float)
			reservations = Reservation.objects.filter(start__gt=form.cleaned_data['start'], start__lte=form.cleaned_data['end'], cancelled=True, cancellation_time__isnull=False)
			if form.cleaned_data['target']:
				reservations = reservations.filter(tool__id=form.cleaned_data['target'])
			for r in reservations:
				cancellation_delta = (r.start - r.cancellation_time).total_seconds()
				if 0 < cancellation_delta < form.cleaned_data['cancellation_horizon']:
					penalty = ((form.cleaned_data['cancellation_horizon'] - cancellation_delta) / form.cleaned_data['cancellation_horizon']) * form.cleaned_data['cancellation_penalty']
					intermediate_results[r.user.id] += penalty
			final_results = {}
			for user_id, score in intermediate_results.items():
				user = User.objects.get(id=user_id)
				final_results[user] = int(score)
			sorted_results = sorted(final_results.items(), key=lambda x: x[1], reverse=True)
			dictionary['results'] = sorted_results
		else:
			form = ReservationAbuseForm()
	except ValidationError:
		form = ReservationAbuseForm()
	dictionary['form'] = form
	return render(request, 'abuse/abuse.html', dictionary)


@staff_member_required(login_url=None)
@require_GET
def user_drill_down(request):
	try:
		form = ReservationAbuseForm(request.GET)
		form.is_valid()
		abuser = User.objects.get(id=request.GET['user'])
		reservations = Reservation.objects.filter(start__gt=form.cleaned_data['start'], start__lte=form.cleaned_data['end'], cancelled=True, cancellation_time__isnull=False, user=abuser)
		if form.cleaned_data['target']:
			reservations = reservations.filter(tool__id=form.cleaned_data['target'])
		abuses = []
		for r in reservations:
			cancellation_delta = (r.start - r.cancellation_time).total_seconds()
			if 0 < cancellation_delta < form.cleaned_data['cancellation_horizon']:
				penalty = ((form.cleaned_data['cancellation_horizon'] - cancellation_delta) / form.cleaned_data['cancellation_horizon']) * form.cleaned_data['cancellation_penalty']
				delta = duration_string((r.start - r.cancellation_time).total_seconds())
				abuses.append({'penalty': penalty, 'start': r.start, 'cancelled': r.cancellation_time, 'delta': delta, 'tool_name': r.tool.name, 'id': r.id})
		return render(request, 'abuse/user_drill_down.html', {'abuses': abuses, 'abuser': abuser})
	except:
		return HttpResponseBadRequest()


def duration_string(duration_in_seconds):
	minutes, seconds = divmod(duration_in_seconds, 60)
	hours, minutes = divmod(minutes, 60)
	hours = int(hours)
	minutes = int(minutes)

	if hours == 0 and minutes == 0:
		return "Less than a minute"

	result = ""

	if hours == 1:
		result = "{0} hour".format(hours)
	elif hours > 1:
		result = "{0} hours".format(hours)

	if minutes == 1:
		if result != "":
			result += ", "
		result += "{0} minute".format(minutes)
	elif minutes > 1:
		if result != "":
			result += ", "
		result += "{0} minutes".format(minutes)

	return result



from unittest import mock

from django.conf import settings
from django.test import TestCase

from NEMO.models import User, Tool, InterlockCard, Interlock, InterlockCardCategory

server1 = 'server1.nist.gov'
server2 = 'https://server2.nist.gov'
port1 = 80
port2 = 8080
ERROR_500 = '500 Server Error'
UNLOCK_ERROR = 'unlock: bad interlock'
LOCK_ERROR = 'lock: bad interlock'

bad_interlock: Interlock = None
wrong_response_interlock: Interlock = None
disabled_interlock: Interlock = None

def web_relay_response(relayNumber, state):
	return f"<datavalues>" \
		   f"  <relay{relayNumber}state>{state}</relay{relayNumber}state>" \
		   f"</datavalues>"


# This method will be used by the mock to replace requests.get
# In the web relay case, it will return a xml file with the relay statuses
def mocked_requests_get(*args, **kwargs):
	from requests import Response
	class MockResponse(Response):

		def __init__(self, content, status_code, reason=None):
			super().__init__()
			self.content = content
			self.status_code = status_code
			self.url = args[0]
			self.reason = reason

		def content(self):
			return self.content

	# interlock 3 on server 2 is bad
	if args[0] == f'{bad_interlock.card.server}:{bad_interlock.card.port}/stateFull.xml?relay3State=1':
		return MockResponse('', 500, UNLOCK_ERROR)
	elif args[0] == f'{bad_interlock.card.server}:{bad_interlock.card.port}/stateFull.xml?relay3State=0':
		return MockResponse('', 500, LOCK_ERROR)
	# interlock 2 on server 2 sends wrong response
	elif args[0] == f'{wrong_response_interlock.card.server}:{wrong_response_interlock.card.port}/stateFull.xml?relay2State=0' or args[0] == f'{wrong_response_interlock.card.server}:{wrong_response_interlock.card.port}/stateFull.xml?relay2State=1':
		return MockResponse('bad response', 200)
	elif 'stateFull.xml' in args[0]:
		url_state = args[0][-8:]
		numbers = [int(i) for i in url_state if i.isdigit()]
		relay_number = numbers[0]
		relay_state = numbers[1]
		return MockResponse(web_relay_response(relay_number, relay_state), 200)

	return MockResponse(None, 404)


class WebRelayInterlockTestCase(TestCase):
	tool: Tool = None

	def setUp(self):
		global tool, wrong_response_interlock, bad_interlock, disabled_interlock
		# enable interlock functionality
		settings.__setattr__('INTERLOCKS_ENABLED', True)
		interlock_card_category = InterlockCardCategory.objects.get(key='web_relay_http')
		interlock_card = InterlockCard.objects.create(server=server1, port=port1, category=interlock_card_category)
		interlock_card2 = InterlockCard.objects.create(server=server2, port=port2, category=interlock_card_category)
		interlock_card3 = InterlockCard.objects.create(server=server2, port=port2, category=interlock_card_category, enabled=False)
		disabled_interlock = Interlock.objects.create(card=interlock_card3, channel=3)
		interlock = Interlock.objects.create(card=interlock_card, channel=1)
		wrong_response_interlock = Interlock.objects.create(card=interlock_card2, channel=2)
		bad_interlock = Interlock.objects.create(card=interlock_card2, channel=3)
		owner = User.objects.create(username='mctest', first_name='Testy', last_name='McTester')
		tool = Tool.objects.create(name='test_tool', primary_owner=owner, interlock=interlock)

	@mock.patch('NEMO.interlocks.requests.get', side_effect=mocked_requests_get)
	def test_disabled_card(self, mock_args):
		self.assertTrue(disabled_interlock.unlock())
		self.assertEquals(disabled_interlock.state, Interlock.State.UNLOCKED)
		self.assertTrue('Interlock interface mocked out' in disabled_interlock.most_recent_reply)
		self.assertTrue(disabled_interlock.lock())
		self.assertEquals(disabled_interlock.state, Interlock.State.LOCKED)
		self.assertTrue('Interlock interface mocked out' in disabled_interlock.most_recent_reply)

	@mock.patch('NEMO.interlocks.requests.get', side_effect=mocked_requests_get)
	def test_all_good(self, mock_args):
		self.assertTrue(tool.interlock.unlock())
		self.assertEquals(tool.interlock.state, Interlock.State.UNLOCKED)
		self.assertTrue(tool.interlock.lock())
		self.assertEquals(tool.interlock.state, Interlock.State.LOCKED)

	@mock.patch('NEMO.interlocks.requests.get', side_effect=mocked_requests_get)
	def test_error_response_from_interlock(self, mock_args):
		self.assertFalse(bad_interlock.unlock())
		self.assertTrue(ERROR_500 in bad_interlock.most_recent_reply)
		self.assertTrue(UNLOCK_ERROR in bad_interlock.most_recent_reply)
		self.assertEquals(bad_interlock.state, Interlock.State.UNKNOWN)

		self.assertFalse(bad_interlock.lock())
		self.assertTrue(ERROR_500 in bad_interlock.most_recent_reply)
		self.assertTrue(LOCK_ERROR in bad_interlock.most_recent_reply)
		self.assertEquals(bad_interlock.state, Interlock.State.UNKNOWN)

	@mock.patch('NEMO.interlocks.requests.get', side_effect=mocked_requests_get)
	def test_wrong_response_from_interlock(self, mock_args):
		self.assertFalse(wrong_response_interlock.unlock())
		self.assertTrue('General exception' in wrong_response_interlock.most_recent_reply)
		self.assertTrue('syntax error' in wrong_response_interlock.most_recent_reply)
		self.assertEquals(wrong_response_interlock.state, Interlock.State.UNKNOWN)



from datetime import timedelta

from django.db.models import Q
from django.http import HttpResponse, HttpResponseBadRequest
from django.template import Template, Context
from django.utils import timezone

from NEMO.models import Reservation, AreaAccessRecord, ScheduledOutage
from NEMO.utilities import format_datetime, send_mail
from NEMO.views.customization import get_customization, get_media_file_contents


def check_policy_to_enable_tool(tool, operator, user, project, staff_charge):
	"""
	Check that the user is allowed to enable the tool. Enable the tool if the policy checks pass.
	"""

	# The tool must be visible (or the parent if it's a child tool) to users.
	visible = tool.parent_tool.visible if tool.is_child_tool() else tool.visible
	if not visible:
		return HttpResponseBadRequest("This tool is currently hidden from users.")

	# The tool must be operational.
	# If the tool is non-operational then it may only be accessed by staff members.
	if not tool.operational and not operator.is_staff:
		return HttpResponseBadRequest("This tool is currently non-operational.")

	# The tool must not be in use.
	current_usage_event = tool.get_current_usage_event()
	if current_usage_event:
		return HttpResponseBadRequest("The tool is currently being used by " + str(current_usage_event.user) + ".")

	# The user must be qualified to use the tool itself, or the parent tool in case of alternate tool.
	tool_to_check_qualifications = tool.parent_tool if tool.is_child_tool() else tool
	if tool_to_check_qualifications not in operator.qualifications.all() and not operator.is_staff:
		return HttpResponseBadRequest("You are not qualified to use this tool.")

	# Only staff members can operate a tool on behalf of another user.
	if (user and operator.pk != user.pk) and not operator.is_staff:
		return HttpResponseBadRequest("You must be a staff member to use a tool on another user's behalf.")

	# All required resources must be available to operate a tool except for staff.
	if tool.required_resource_set.filter(available=False).exists() and not operator.is_staff:
		return HttpResponseBadRequest("A resource that is required to operate this tool is unavailable.")

	# The tool operator may not activate tools in a particular area unless they are logged in to the area.
	# Staff are exempt from this rule.
	if tool.requires_area_access and AreaAccessRecord.objects.filter(area=tool.requires_area_access, customer=operator, staff_charge=None, end=None).count() == 0 and not operator.is_staff:
		dictionary = {
			'operator': operator,
			'tool': tool,
		}
		abuse_email_address = get_customization('abuse_email_address')
		message = get_media_file_contents('unauthorized_tool_access_email.html')
		if abuse_email_address and message:
			rendered_message = Template(message).render(Context(dictionary))
			send_mail("Area access requirement", rendered_message, abuse_email_address, [abuse_email_address])
		return HttpResponseBadRequest("You must be logged in to the {} to operate this tool.".format(tool.requires_area_access.name.lower()))

	# Staff may only charge staff time for one user at a time.
	if staff_charge and operator.charging_staff_time():
		return HttpResponseBadRequest('You are already charging staff time. You must end the current staff charge before you being another.')

	# Staff may not bill staff time to the themselves.
	if staff_charge and operator == user:
		return HttpResponseBadRequest('You cannot charge staff time to yourself.')

	# Users may only charge to projects they are members of.
	if project not in user.active_projects():
		return HttpResponseBadRequest('The designated user is not assigned to the selected project.')

	# The tool operator must not have a lock on usage
	if operator.training_required:
		return HttpResponseBadRequest("You are blocked from using all tools in the NanoFab. Please complete the NanoFab rules tutorial in order to use tools.")

	# Users may only use a tool when delayed logoff is not in effect. Staff are exempt from this rule.
	if tool.delayed_logoff_in_progress() and not operator.is_staff:
		return HttpResponseBadRequest("Delayed tool logoff is in effect. You must wait for the delayed logoff to expire before you can use the tool.")

	# Users may not enable a tool during a scheduled outage. Staff are exempt from this rule.
	if tool.scheduled_outage_in_progress() and not operator.is_staff:
		return HttpResponseBadRequest("A scheduled outage is in effect. You must wait for the outage to end before you can use the tool.")

	return HttpResponse()


def check_policy_to_disable_tool(tool, operator, downtime):
	""" Check that the user is allowed to disable the tool. """
	current_usage_event = tool.get_current_usage_event()
	if current_usage_event.operator != operator and current_usage_event.user != operator and not operator.is_staff:
		return HttpResponseBadRequest('You may not disable a tool while another user is using it unless you are a staff member.')
	if downtime < timedelta():
		return HttpResponseBadRequest('Downtime cannot be negative.')
	if downtime > timedelta(minutes=120):
		return HttpResponseBadRequest('Post-usage tool downtime may not exceed 120 minutes.')
	if tool.delayed_logoff_in_progress() and downtime > timedelta():
		return HttpResponseBadRequest('The tool is already in a delayed-logoff state. You may not issue additional delayed logoffs until the existing one expires.')
	if not tool.allow_delayed_logoff and downtime > timedelta():
		return HttpResponseBadRequest('Delayed logoff is not allowed for this tool.')
	return HttpResponse()


def check_policy_to_save_reservation(cancelled_reservation, new_reservation, user, explicit_policy_override):
	""" Check the reservation creation policy and return a list of policy problems """

	# The function will check all policies. Policy problems are placed in the policy_problems list. overridable is True if the policy problems can be overridden by a staff member.
	policy_problems = []
	overridable = False

	# Reservations may not have a start time that is earlier than the end time.
	if new_reservation.start >= new_reservation.end:
		policy_problems.append("Reservation start time (" + format_datetime(new_reservation.start) + ") must be before the end time (" + format_datetime(new_reservation.end) + ").")

	# The user may not create, move, or resize a reservation to coincide with another user's reservation.
	coincident_events = Reservation.objects.filter(tool=new_reservation.tool, cancelled=False, missed=False, shortened=False)
	# Exclude the reservation we're cancelling in order to create a new one:
	if cancelled_reservation and cancelled_reservation.id:
		coincident_events = coincident_events.exclude(id=cancelled_reservation.id)
	# Exclude events for which the following is true:
	# The event starts and ends before the time-window, and...
	# The event starts and ends after the time-window.
	coincident_events = coincident_events.exclude(start__lt=new_reservation.start, end__lte=new_reservation.start)
	coincident_events = coincident_events.exclude(start__gte=new_reservation.end, end__gt=new_reservation.end)
	if coincident_events.count() > 0:
		policy_problems.append("Your reservation coincides with another reservation that already exists. Please choose a different time.")

	# The user may not create, move, or resize a reservation to coincide with a scheduled outage.
	coincident_events = ScheduledOutage.objects.filter(Q(tool=new_reservation.tool) | Q(resource__fully_dependent_tools__in=[new_reservation.tool]))
	# Exclude events for which the following is true:
	# The event starts and ends before the time-window, and...
	# The event starts and ends after the time-window.
	coincident_events = coincident_events.exclude(start__lt=new_reservation.start, end__lte=new_reservation.start)
	coincident_events = coincident_events.exclude(start__gte=new_reservation.end, end__gt=new_reservation.end)
	if coincident_events.count() > 0:
		policy_problems.append("Your reservation coincides with a scheduled outage. Please choose a different time.")

	# Reservations that have been cancelled may not be changed.
	if new_reservation.cancelled:
		policy_problems.append("This reservation has already been cancelled by " + str(new_reservation.cancelled_by) + " at " + format_datetime(new_reservation.cancellation_time) + ".")

	# The user must belong to at least one active project to make a reservation.
	if new_reservation.user.active_project_count() < 1:
		if new_reservation.user == user:
			policy_problems.append("You do not belong to any active projects. Thus, you may not create any reservations.")
		else:
			policy_problems.append(str(new_reservation.user) + " does not belong to any active projects and cannot have reservations.")

	# The user must associate their reservation with a project they belong to.
	if new_reservation.project and new_reservation.project not in new_reservation.user.active_projects():
		if new_reservation.user == user:
			policy_problems.append("You do not belong to the project associated with this reservation.")
		else:
			policy_problems.append(str(new_reservation.user) + " does not belong to the project named " + str(new_reservation.project) + ".")

	# If the user is a staff member or there's an explicit policy override then the policy check is finished.
	if user.is_staff or explicit_policy_override:
		return policy_problems, overridable

	# If there are no blocking policy conflicts at this point, the rest of the policies can be overridden.
	if not policy_problems:
		overridable = True

	# The user must complete NEMO training to create reservations.
	# Staff may break this rule.
	# An explicit policy override allows this rule to be broken.
	if user.training_required:
		policy_problems.append("You are blocked from making reservations for all tools in the NanoFab. Please complete the NanoFab rules tutorial in order to create new reservations.")

	# Users may only change their own reservations.
	# Staff may break this rule.
	# An explicit policy override allows this rule to be broken.
	if new_reservation.user != user:
		policy_problems.append("You may not change reservations that you do not own.")

	# The user may not create or move a reservation to have a start time that is earlier than the current time.
	# Staff may break this rule.
	# An explicit policy override allows this rule to be broken.
	if new_reservation.start < timezone.now():
		policy_problems.append("Reservation start time (" + format_datetime(new_reservation.start) + ") is earlier than the current time (" + format_datetime(timezone.now()) + ").")

	# The user may not move or resize a reservation to have an end time that is earlier than the current time.
	# Staff may break this rule.
	# An explicit policy override allows this rule to be broken.
	if new_reservation.end < timezone.now():
		policy_problems.append("Reservation end time (" + format_datetime(new_reservation.end) + ") is earlier than the current time (" + format_datetime(timezone.now()) + ").")

	# The user must be qualified on the tool in question in order to create, move, or resize a reservation.
	# Staff may break this rule.
	# An explicit policy override allows this rule to be broken.
	if new_reservation.tool not in user.qualifications.all():
		policy_problems.append("You are not qualified to use this tool. Creating, moving, and resizing reservations is forbidden.")

	# The reservation start time may not exceed the tool's reservation horizon.
	# Staff may break this rule.
	# An explicit policy override allows this rule to be broken.
	if new_reservation.tool.reservation_horizon is not None:
		reservation_horizon = timedelta(days=new_reservation.tool.reservation_horizon)
		if new_reservation.start > timezone.now() + reservation_horizon:
			policy_problems.append("You may not create reservations further than " + str(reservation_horizon.days) + " days from now for this tool.")


	# Check tool policy rules
	tool_policy_problems = []
	if new_reservation.tool.should_enforce_policy(new_reservation):
		tool_policy_problems = check_policy_rules_for_tool(cancelled_reservation, new_reservation, user)

	# Return the list of all policies that are not met.
	return policy_problems + tool_policy_problems, overridable


def check_policy_rules_for_tool(cancelled_reservation, new_reservation, user):
	tool_policy_problems = []
	# Calculate the duration of the reservation:
	duration = new_reservation.end - new_reservation.start

	# The reservation must be at least as long as the minimum block time for this tool.
	# Staff may break this rule.
	# An explicit policy override allows this rule to be broken.
	if new_reservation.tool.minimum_usage_block_time:
		minimum_block_time = timedelta(minutes=new_reservation.tool.minimum_usage_block_time)
		if duration < minimum_block_time:
			tool_policy_problems.append("Your reservation has a duration of " + str(int(
				duration.total_seconds() / 60)) + " minutes. This tool requires a minimum reservation duration of " + str(
				int(minimum_block_time.total_seconds() / 60)) + " minutes.")

	# The reservation may not exceed the maximum block time for this tool.
	# Staff may break this rule.
	# An explicit policy override allows this rule to be broken.
	if new_reservation.tool.maximum_usage_block_time:
		maximum_block_time = timedelta(minutes=new_reservation.tool.maximum_usage_block_time)
		if duration > maximum_block_time:
			tool_policy_problems.append("Your reservation has a duration of " + str(
				int(duration.total_seconds() / 60)) + " minutes. Reservations for this tool may not exceed " + str(
				int(maximum_block_time.total_seconds() / 60)) + " minutes.")

	# If there is a limit on number of reservations per user per day then verify that the user has not exceeded it.
	# Staff may break this rule.
	# An explicit policy override allows this rule to be broken.
	if new_reservation.tool.maximum_reservations_per_day:
		start_of_day = new_reservation.start
		start_of_day = start_of_day.replace(hour=0, minute=0, second=0, microsecond=0)
		end_of_day = start_of_day + timedelta(days=1)
		reservations_for_that_day = Reservation.objects.filter(cancelled=False, shortened=False,
															   start__gte=start_of_day, end__lte=end_of_day, user=user,
															   tool=new_reservation.tool)
		# Exclude any reservation that is being cancelled.
		if cancelled_reservation and cancelled_reservation.id:
			reservations_for_that_day = reservations_for_that_day.exclude(id=cancelled_reservation.id)
		if reservations_for_that_day.count() >= new_reservation.tool.maximum_reservations_per_day:
			tool_policy_problems.append("You may only have " + str(
				new_reservation.tool.maximum_reservations_per_day) + " reservations for this tool per day. Missed reservations are included when counting the number of reservations per day.")

	# A minimum amount of time between reservations for the same user & same tool can be enforced.
	# Staff may break this rule.
	# An explicit policy override allows this rule to be broken.
	if new_reservation.tool.minimum_time_between_reservations:
		buffer_time = timedelta(minutes=new_reservation.tool.minimum_time_between_reservations)
		must_end_before = new_reservation.start - buffer_time
		too_close = Reservation.objects.filter(cancelled=False, shortened=False, user=user, end__gt=must_end_before,
											   start__lt=new_reservation.start, tool=new_reservation.tool)
		if cancelled_reservation and cancelled_reservation.id:
			too_close = too_close.exclude(id=cancelled_reservation.id)
		if too_close.exists():
			tool_policy_problems.append("Separate reservations for this tool that belong to you must be at least " + str(
				new_reservation.tool.minimum_time_between_reservations) + " minutes apart from each other. The proposed reservation ends too close to another reservation.")
		must_start_after = new_reservation.end + buffer_time
		too_close = Reservation.objects.filter(cancelled=False, shortened=False, user=user, start__lt=must_start_after,
											   end__gt=new_reservation.start, tool=new_reservation.tool)
		if cancelled_reservation and cancelled_reservation.id:
			too_close = too_close.exclude(id=cancelled_reservation.id)
		if too_close.exists():
			tool_policy_problems.append("Separate reservations for this tool that belong to you must be at least " + str(
				new_reservation.tool.minimum_time_between_reservations) + " minutes apart from each other. The proposed reservation begins too close to another reservation.")

	# Check that the user is not exceeding the maximum amount of time they may reserve in the future.
	# Staff may break this rule.
	# An explicit policy override allows this rule to be broken.
	if new_reservation.tool.maximum_future_reservation_time:
		reservations_after_now = Reservation.objects.filter(cancelled=False, user=user, tool=new_reservation.tool,
															start__gte=timezone.now())
		if cancelled_reservation and cancelled_reservation.id:
			reservations_after_now = reservations_after_now.exclude(id=cancelled_reservation.id)
		amount_reserved_in_the_future = new_reservation.duration()
		for r in reservations_after_now:
			amount_reserved_in_the_future += r.duration()
		if amount_reserved_in_the_future.total_seconds() / 60 > new_reservation.tool.maximum_future_reservation_time:
			tool_policy_problems.append("You may only reserve up to " + str(
				new_reservation.tool.maximum_future_reservation_time) + " minutes of time on this tool, starting from the current time onward.")

	return tool_policy_problems



def check_policy_to_cancel_reservation(reservation, user):
	"""
	Checks the reservation deletion policy.
	If all checks pass the function returns an HTTP "OK" response.
	Otherwise, the function returns an HTTP "Bad Request" with an error message.
	"""

	# Users may only cancel reservations that they own.
	# Staff may break this rule.
	if (reservation.user != user) and not user.is_staff:
		return HttpResponseBadRequest("You may not cancel reservations that you do not own.")

	# Users may not cancel reservations that have already ended.
	# Staff may break this rule.
	if reservation.end < timezone.now() and not user.is_staff:
		return HttpResponseBadRequest("You may not cancel reservations that have already ended.")

	if reservation.cancelled:
		return HttpResponseBadRequest("This reservation has already been cancelled by " + str(reservation.cancelled_by) + " at " + format_datetime(reservation.cancellation_time) + ".")

	if reservation.missed:
		return HttpResponseBadRequest("This reservation was missed and cannot be modified.")

	return HttpResponse()


def check_policy_to_create_outage(outage):
	policy_problems = []
	# Outages may not have a start time that is earlier than the end time.
	if outage.start >= outage.end:
		return "Outage start time (" + format_datetime(outage.start) + ") must be before the end time (" + format_datetime(outage.end) + ")."

	# The user may not create, move, or resize an outage to coincide with another user's reservation.
	coincident_events = Reservation.objects.filter(tool=outage.tool, cancelled=False, missed=False, shortened=False)
	# Exclude events for which the following is true:
	# The event starts and ends before the time-window, and...
	# The event starts and ends after the time-window.
	coincident_events = coincident_events.exclude(start__lt=outage.start, end__lte=outage.start)
	coincident_events = coincident_events.exclude(start__gte=outage.end, end__gt=outage.end)
	if coincident_events.count() > 0:
		return "Your scheduled outage coincides with a reservation that already exists. Please choose a different time."

	# No policy issues! The outage can be created...
	return None



from setuptools import setup, find_packages

setup(
	name='NEMO',
	version='1.22.0',
	python_requires='>=3.6',
	packages=find_packages(exclude=['NEMO.tests']),
	include_package_data=True,
	url='https://github.com/usnistgov/NEMO',
	license='Public domain',
	author='Center for Nanoscale Science and Technology',
	author_email='CNSTapplications@nist.gov',
	description='NEMO is a laboratory logistics web application. Use it to schedule reservations, control tool access, track maintenance issues, and more.',
	long_description='Find out more about NEMO on the GitHub project page https://github.com/usnistgov/NEMO',
	classifiers=[
		'Development Status :: 5 - Production/Stable',
		'Environment :: Web Environment',
		'Framework :: Django',
		'Intended Audience :: Science/Research',
		'Intended Audience :: System Administrators',
		'License :: Public Domain',
		'Natural Language :: English',
		'Operating System :: OS Independent',
		'Programming Language :: Python :: 3.6',
	],
	install_requires=[
		'cryptography==2.3.1',
		'Django==1.11.27',
		'django-filter==1.1.0',
		'djangorestframework==3.9.4',
		'ldap3==2.5.1',
		'python-dateutil==2.7.3',
		'requests==2.21.0',
		'Pillow==6.2.0',
	],
	entry_points={
		'console_scripts': ['nemo=NEMO.provisioning:entry_point'],
	},
)
